import h5py
import tensorflow as tf

num_feature = 5

def parse_proto(example_proto):
    features = {
        # 'X': tf.FixedLenFeature((num_feature,), tf.float32, allow_missing = True)
        # 'X': tf.FixedLenFeature((num_feature,), tf.float32)
        'X': tf.compat.v1.FixedLenFeature([num_feature], tf.float32)
        # 'X': tf.FixedLenFeature([], tf.float32)
    }
    # parsed_features = tf.compat.v1.parse_single_example(example_proto, features)
    parsed_features = tf.io.parse_single_example(serialized = example_proto,
                                                        features   = features)
    return([parsed_features['X']])

class parser():
    """ """
    def __init__(self, features = {}):
        self.features = features

    def parse(self, example):
        parsed_features = tf.io.parse_single_example(serialized = example,
                                                     features   = self.features)
        return([value for key,value in parsed_features.items()])



import os
import time
import logging

import numpy as np

from sklearn.neighbors import KNeighborsClassifier
from sklearn.linear_model import LogisticRegression
from sklearn.neural_network import MLPClassifier

from .data.data_handler import kfold_split


def run_classifier(X_train, X_test, y_train, y_test, classifier = "KNeighborsClassifier", max_dim = 1e6, args = {}):
    if classifier == "KNeighborsClassifier":
        clf = KNeighborsClassifier(**args)
    elif classifier == "LogisticRegression":
        clf = LogisticRegression(**args)
    elif classifier == "MLPClassifier":
        clf = MLPClassifier(**args)
    else:
        raise Exception("Input valid classifier")
    _ = clf.fit(X_train,y_train)
    score = clf.score(X_test,y_test)
    return score, clf


# classifier.run_classifiers_over_params(X_classifier,y,k_split,clf_types,classifier_args_dict,classifier_args_ps_dict)

def run_classifiers_over_params(X,y, k_split, clf_types,
                                classifier_args_dict, classifier_args_ps_dict,
                                logdir = "", random_seed = 0, save_npy = False):
    """
    X: matrix
    y: label
    data_splits:
    """

    data_splits = kfold_split(X, y, k_split, random_seed)

    scores = []

    for i_clf, clf_type in enumerate(clf_types):
        print("")
        print("================ Running {} ======================".format(clf_type))

        args = dict(classifier_args_dict[clf_type])

        parameter_dict = classifier_args_ps_dict[clf_type]

        for param_name, param_list in parameter_dict.items():

            npy_scores_param = os.path.join(logdir,"scores_{}.npy".format(clf_type))

            if os.path.isfile(npy_scores_param) and False:
                scores_param = np.load(npy_scores_param)
            else:
                ## Choose the best param
                scores_param = []

                for i_param, param in enumerate(param_list):
                    print("{} = {}".format(param_name, param))
                    args[param_name] = param
                    scores_kfold = np.zeros(len(data_splits))
                    clf_best = None

                    for i_kfold, data_split in enumerate(data_splits):
                        # print("kfold = {}".format(i_kfold))
                        X_train, X_test, y_train, y_test = data_split
                        score, clf = run_classifier(X_train = X_train,
                                                    X_test = X_test,
                                                    y_train = y_train,
                                                    y_test = y_test,
                                                    classifier = clf_type,
                                                    args = args)
                        scores_kfold[i_kfold] = score

                    score_kfold = scores_kfold.mean()
                    scores_param.append(scores_kfold)

                scores_param = np.array(scores_param)
                score_best_param = scores_param[np.argmax(scores_param.mean(-1))]

                if save_npy:
                    np.save(npy_scores_param, np.array(score_best_param))

            print(score_best_param)

        scores.append(score_best_param)

    return scores



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import sys
import numpy as np
from skimage.util import view_as_windows
import warnings, logging
import tensorflow as tf
from tensorflow.python.framework import ops
from tensorflow.python.ops import nn_grad, math_grad
from collections import OrderedDict
from .utils import make_batches, slice_arrays, to_list, unpack_singleton, placeholder_from_data

SUPPORTED_ACTIVATIONS = [
    'Relu', 'Elu', 'Sigmoid', 'Tanh', 'Softplus'
]

UNSUPPORTED_ACTIVATIONS = [
    'CRelu', 'Relu6', 'Softsign'
]

_ENABLED_METHOD_CLASS = None
_GRAD_OVERRIDE_CHECKFLAG = 0

# -----------------------------------------------------------------------------
# CUSTOM UTILITY FUNCTIONS for identifying input specific operations in DeepLIFT
# -----------------------------------------------------------------------------

from itertools import chain

def children(op):
  return set(op for out in op.outputs for op in out.consumers())


def descendants(op, depth = 0, max_depth = 100):
    # print(op.name)
    ops_children = children(op)
    if len(ops_children) == 0 or depth >= max_depth:
        ops_descendants = set([op])
    else:
        ops_descendants = set(chain.from_iterable([descendants(op_child, depth = depth + 1) for op_child in ops_children]))
        ops_descendants = set.union(ops_descendants, ops_children)
    return ops_descendants


# -----------------------------------------------------------------------------
# UTILITY FUNCTIONS
# -----------------------------------------------------------------------------


def activation(type):
    """
    Returns Tensorflow's activation op, given its type
    :param type: string
    :return: op
    """
    if type not in SUPPORTED_ACTIVATIONS:
        warnings.warn('Activation function (%s) not supported' % type)
    f = getattr(tf.nn, type.lower())
    return f


def original_grad(op, grad):
    """
    Return original Tensorflow gradient for an op
    :param op: op
    :param grad: Tensor
    :return: Tensor
    """
    if op.type not in SUPPORTED_ACTIVATIONS:
        warnings.warn('Activation function (%s) not supported' % op.type)
    opname = '_%sGrad' % op.type
    if hasattr(nn_grad, opname):
        f = getattr(nn_grad, opname)
    else:
        f = getattr(math_grad, opname)
    return f(op, grad)


# -----------------------------------------------------------------------------
# ATTRIBUTION METHODS BASE CLASSES
# -----------------------------------------------------------------------------


class AttributionMethod(object):
    """
    Attribution method base class
    """
    def __init__(self, T, X, session, keras_learning_phase=None, feed_dict = {}):
        self.T = T  # target Tensor
        self.X = X  # input Tensor
        self.Y_shape = [None,] + T.get_shape().as_list()[1:]
        # Most often T contains multiple output units. In this case, it is often necessary to select
        # a single unit to compute contributions for. This can be achieved passing 'ys' as weight for the output Tensor.
        self.Y = tf.placeholder(tf.float32, self.Y_shape)
        # placeholder_from_data(ys) if ys is not None else 1.0  # Tensor that represents weights for T
        self.T = self.T * self.Y
        self.feed_dict = feed_dict
        self.symbolic_attribution = None
        self.session = session
        self.keras_learning_phase = keras_learning_phase
        self.has_multiple_inputs = type(self.X) is list or type(self.X) is tuple
        logging.info('Model with multiple inputs: %s' % self.has_multiple_inputs)

        # Set baseline
        # TODO: now this sets a baseline also for those methods that does not require it
        self._set_check_baseline()

        # References
        self._init_references()

        # Create symbolic explanation once during construction (affects only gradient-based methods)
        self.explain_symbolic()

    def explain_symbolic(self):
        return None

    def run(self, xs, ys=None, batch_size=None):
        pass

    def _init_references(self):
        pass

    def _check_input_compatibility(self, xs, ys=None, batch_size=None):
        if ys is not None:
            if not self.has_multiple_inputs and len(xs) != len(ys):
                raise RuntimeError('When provided, ys must have the same batch size as xs (xs has batch size {} and ys {})'.format(len(xs), len(ys)))
            elif self.has_multiple_inputs and np.all([len(i) != len(ys) for i in xs]):
                raise RuntimeError('When provided, ys must have the same batch size as all elements of xs')
        if batch_size is not None and batch_size > 0:
            if self.T.shape[0].value is not None and self.T.shape[0].value is not batch_size:
                raise RuntimeError('When using batch evaluation, the first dimension of the target tensor '
                                   'must be compatible with the batch size. Found %s instead' % self.T.shape[0].value)
            if isinstance(self.X, list):
                for x in self.X:
                    if x.shape[0].value is not None and x.shape[0].value is not batch_size:
                        raise RuntimeError('When using batch evaluation, the first dimension of the input tensor '
                                           'must be compatible with the batch size. Found %s instead' % x.shape[
                                               0].value)
            else:
                if self.X.shape[0].value is not None and self.X.shape[0].value is not batch_size:
                    raise RuntimeError('When using batch evaluation, the first dimension of the input tensor '
                                       'must be compatible with the batch size. Found %s instead' % self.X.shape[0].value)

    def _session_run_batch(self, T, xs, ys=None, feed_dict = {}):

        if self.has_multiple_inputs:
            for k, v in zip(self.X, xs):
                feed_dict[k] = v
        else:
            feed_dict[self.X] = xs

        # If ys is not passed, produce a vector of ones that will be broadcasted to all batch samples
        feed_dict[self.Y] = ys if ys is not None else np.ones([1,] + self.Y_shape[1:])

        if self.keras_learning_phase is not None:
            feed_dict[self.keras_learning_phase] = 0
        return self.session.run(T, feed_dict)

    def _session_run(self, T, xs, ys=None, batch_size=None):
        num_samples = len(xs)
        if self.has_multiple_inputs is True:
            num_samples = len(xs[0])
            if len(xs) != len(self.X):
                raise RuntimeError('List of input tensors and input data have different lengths (%s and %s)'
                                   % (str(len(xs)), str(len(self.X))))
            if batch_size is not None:
                for xi in xs:
                    if len(xi) != num_samples:
                        raise RuntimeError('Evaluation in batches requires all inputs to have '
                                           'the same number of samples')

        if batch_size is None or batch_size <= 0 or num_samples <= batch_size:
            return self._session_run_batch(T, xs, ys, feed_dict = self.feed_dict)
        else:
            outs = []
            batches = make_batches(num_samples, batch_size)
            for batch_index, (batch_start, batch_end) in enumerate(batches):
                # Get a batch from data
                xs_batch = slice_arrays(xs, batch_start, batch_end)
                # If the target tensor has one entry for each sample, we need to batch it as well
                ys_batch = None
                if ys is not None:
                    ys_batch = slice_arrays(ys, batch_start, batch_end)
                batch_outs = self._session_run_batch(T, xs_batch, ys_batch, feed_dict = self.feed_dict)
                batch_outs = to_list(batch_outs)
                if batch_index == 0:
                    # Pre-allocate the results arrays.
                    for batch_out in batch_outs:
                        shape = (num_samples,) + batch_out.shape[1:]
                        outs.append(np.zeros(shape, dtype=batch_out.dtype))
                for i, batch_out in enumerate(batch_outs):
                    outs[i][batch_start:batch_end] = batch_out
            return unpack_singleton(outs)

    def _set_check_baseline(self):
        # Do nothing for those methods that have no baseline required
        if not hasattr(self, "baseline"):
            return

        if self.baseline is None:
            if self.has_multiple_inputs:
                self.baseline = [np.zeros([1,] + xi.get_shape().as_list()[1:]) for xi in self.X]
            else:
                self.baseline = np.zeros([1,] + self.X.get_shape().as_list()[1:])

        else:
            if self.has_multiple_inputs:
                for i, xi in enumerate(self.X):
                    if list(self.baseline[i].shape) == xi.get_shape().as_list()[1:]:
                        self.baseline[i] = np.expand_dims(self.baseline[i], 0)
                    else:
                        raise RuntimeError('Baseline shape %s does not match expected shape %s'
                                           % (self.baseline[i].shape, self.X.get_shape().as_list()[1:]))
            else:
                if list(self.baseline.shape) == self.X.get_shape().as_list()[1:]:
                    self.baseline = np.expand_dims(self.baseline, 0)
                else:
                    raise RuntimeError('Baseline shape %s does not match expected shape %s'
                                       % (self.baseline.shape, self.X.get_shape().as_list()[1:]))


class GradientBasedMethod(AttributionMethod):
    """
    Base class for gradient-based attribution methods
    """
    def get_symbolic_attribution(self):
        return tf.gradients(self.T, self.X)

    def explain_symbolic(self):
        if self.symbolic_attribution is None:
            self.symbolic_attribution = self.get_symbolic_attribution()
        return self.symbolic_attribution

    def run(self, xs, ys=None, batch_size=None):
        self._check_input_compatibility(xs, ys, batch_size)
        results = self._session_run(self.explain_symbolic(), xs, ys, batch_size)
        return results[0] if not self.has_multiple_inputs else results

    @classmethod
    def nonlinearity_grad_override(cls, op, grad):
        return original_grad(op, grad)


class PerturbationBasedMethod(AttributionMethod):
    """
       Base class for perturbation-based attribution methods
       """
    def __init__(self, T, X, session, keras_learning_phase, feed_dict):
        super(PerturbationBasedMethod, self).__init__(T, X, session, keras_learning_phase, feed_dict)
        self.base_activation = None



# -----------------------------------------------------------------------------
# ATTRIBUTION METHODS
# -----------------------------------------------------------------------------
"""
Returns zero attributions. For testing only.
"""


class DummyZero(GradientBasedMethod):

    def get_symbolic_attribution(self,):
        return tf.gradients(self.T, self.X)

    @classmethod
    def nonlinearity_grad_override(cls, op, grad):
        input = op.inputs[0]
        return tf.zeros_like(input)

"""
Saliency maps
https://arxiv.org/abs/1312.6034
"""


class Saliency(GradientBasedMethod):

    def get_symbolic_attribution(self):
        return [tf.abs(g) for g in tf.gradients(self.T, self.X)]


"""
Gradient * Input
https://arxiv.org/pdf/1704.02685.pdf - https://arxiv.org/abs/1611.07270
"""


class GradientXInput(GradientBasedMethod):

    def get_symbolic_attribution(self):
        return [g * x for g, x in zip(
            tf.gradients(self.T, self.X),
            self.X if self.has_multiple_inputs else [self.X])]


"""
Integrated Gradients
https://arxiv.org/pdf/1703.01365.pdf
"""


class IntegratedGradients(GradientBasedMethod):

    def __init__(self, T, X, session, keras_learning_phase, feed_dict, steps=100, baseline=None):
        self.steps = steps
        self.baseline = baseline
        super(IntegratedGradients, self).__init__(T, X, session, keras_learning_phase, feed_dict)

    def run(self, xs, ys=None, batch_size=None):
        self._check_input_compatibility(xs, ys, batch_size)

        gradient = None
        for alpha in list(np.linspace(1. / self.steps, 1.0, self.steps)):
            xs_mod = [b + (x - b) * alpha for x, b in zip(xs, self.baseline)] if self.has_multiple_inputs \
                else self.baseline + (xs - self.baseline) * alpha
            _attr = self._session_run(self.explain_symbolic(), xs_mod, ys, batch_size)
            if gradient is None: gradient = _attr
            else: gradient = [g + a for g, a in zip(gradient, _attr)]

        results = [g * (x - b) / self.steps for g, x, b in zip(
            gradient,
            xs if self.has_multiple_inputs else [xs],
            self.baseline if self.has_multiple_inputs else [self.baseline])]

        return results[0] if not self.has_multiple_inputs else results


"""
Layer-wise Relevance Propagation with epsilon rule
http://journals.plos.org/plosone/article?id=10.1371/journal.pone.0130140
"""


class EpsilonLRP(GradientBasedMethod):
    eps = None

    def __init__(self, T, X, session, keras_learning_phase, feed_dict, epsilon=1e-4):
        assert epsilon > 0.0, 'LRP epsilon must be greater than zero'
        global eps
        eps = epsilon
        super(EpsilonLRP, self).__init__(T, X, session, keras_learning_phase, feed_dict)

    def get_symbolic_attribution(self):
        return [g * x for g, x in zip(
            tf.gradients(self.T, self.X),
            self.X if self.has_multiple_inputs else [self.X])]

    @classmethod
    def nonlinearity_grad_override(cls, op, grad):
        output = op.outputs[0]
        input = op.inputs[0]
        return grad * output / (input + eps *
                                tf.where(input >= 0, tf.ones_like(input), -1 * tf.ones_like(input)))

"""
DeepLIFT
This reformulation only considers the "Rescale" rule
https://arxiv.org/abs/1704.02685
"""


class DeepLIFTRescale(GradientBasedMethod):

    _deeplift_ref = {}

    def __init__(self, T, X, session, keras_learning_phase, feed_dict, baseline=None, init_ref = 'default'):
        self.baseline = baseline
        self.X = X
        self.init_ref = init_ref
        super(DeepLIFTRescale, self).__init__(T, X, session, keras_learning_phase, feed_dict)

    def get_symbolic_attribution(self):
        return [g * (x - b) for g, x, b in zip(
            tf.gradients(self.T, self.X),
            self.X if self.has_multiple_inputs else [self.X],
            self.baseline if self.has_multiple_inputs else [self.baseline])]

    @classmethod
    def nonlinearity_grad_override(cls, op, grad):
        output = op.outputs[0]
        input = op.inputs[0]
        ref_input = cls._deeplift_ref[op.name]
        ref_output = activation(op.type)(ref_input)
        delta_out = output - ref_output
        delta_in = input - ref_input
        instant_grad = activation(op.type)(0.5 * (ref_input + input))
        return tf.where(tf.abs(delta_in) > 1e-5, grad * delta_out / delta_in,
                        original_grad(instant_grad.op, grad))

    def _init_references(self):
        # print ('DeepLIFT: computing references...')
        sys.stdout.flush()
        self._deeplift_ref.clear()
        ops = []

        # Original: Full set of operations
        if self.init_ref is 'default':
            g = tf.get_default_graph()
            ops_check = g.get_operations()
        elif self.init_ref is 'custom':
            ## Custom: limit operations to those downstream of the input
            print('Custom: limit operations to those downstream of the input')
            ops_check = [descendants(op) for op in self.X.consumers()]
            ops_check = list(set(chain.from_iterable(ops_check)))

        for op in ops_check:
            if len(op.inputs) > 0 and not op.name.startswith('gradients'):
                if op.type in SUPPORTED_ACTIVATIONS:
                    ops.append(op)

        YR = self._session_run([o.inputs[0] for o in ops], self.baseline)
        for (r, op) in zip(YR, ops):
            self._deeplift_ref[op.name] = r
        # print('DeepLIFT: references ready')
        sys.stdout.flush()


"""
Occlusion method
Generalization of the grey-box method presented in https://arxiv.org/pdf/1311.2901.pdf
This method performs a systematic perturbation of contiguous hyperpatches in the input,
replacing each patch with a user-defined value (by default 0).

window_shape : integer or tuple of length xs_ndim
Defines the shape of the elementary n-dimensional orthotope the rolling window view.
If an integer is given, the shape will be a hypercube of sidelength given by its value.

step : integer or tuple of length xs_ndim
Indicates step size at which extraction shall be performed.
If integer is given, then the step is uniform in all dimensions.
"""


class Occlusion(PerturbationBasedMethod):

    def __init__(self, T, X, session, keras_learning_phase, feed_dict, window_shape=None, step=None):
        super(Occlusion, self).__init__(T, X, session, keras_learning_phase, feed_dict)
        if self.has_multiple_inputs:
            raise RuntimeError('Multiple inputs not yet supported for perturbation methods')

        input_shape = X[0].get_shape().as_list()
        if window_shape is not None:
            assert len(window_shape) == len(input_shape), \
                'window_shape must have length of input (%d)' % len(input_shape)
            self.window_shape = tuple(window_shape)
        else:
            self.window_shape = (1,) * len(input_shape)

        if step is not None:
            assert isinstance(step, int) or len(step) == len(input_shape), \
                'step must be integer or tuple with the length of input (%d)' % len(input_shape)
            self.step = step
        else:
            self.step = 1
        self.replace_value = 0.0
        logging.info('Input shape: %s; window_shape %s; step %s' % (input_shape, self.window_shape, self.step))

    def run(self, xs, ys=None, batch_size=None):
        self._check_input_compatibility(xs, ys, batch_size)
        input_shape = xs.shape[1:]
        batch_size = xs.shape[0]
        total_dim = np.asscalar(np.prod(input_shape))

        # Create mask
        index_matrix = np.arange(total_dim).reshape(input_shape)
        idx_patches = view_as_windows(index_matrix, self.window_shape, self.step).reshape((-1,) + self.window_shape)
        heatmap = np.zeros_like(xs, dtype=np.float32).reshape((-1), total_dim)
        w = np.zeros_like(heatmap)

        # Compute original output
        eval0 = self._session_run(self.T, xs, ys, batch_size)

        # Start perturbation loop
        for i, p in enumerate(idx_patches):
            mask = np.ones(input_shape).flatten()
            mask[p.flatten()] = self.replace_value
            masked_xs = mask.reshape((1,) + input_shape) * xs
            delta = eval0 - self._session_run(self.T, masked_xs, ys, batch_size)
            delta_aggregated = np.sum(delta.reshape((batch_size, -1)), -1, keepdims=True)
            heatmap[:, p.flatten()] += delta_aggregated
            w[:, p.flatten()] += p.size

        attribution = np.reshape(heatmap / w, xs.shape)
        if np.isnan(attribution).any():
            warnings.warn('Attributions generated by Occlusion method contain nans, '
                          'probably because window_shape and step do not allow to cover the all input.')
        return attribution


"""
Shapley Value sampling
Computes approximate Shapley Values using "Polynomial calculation of the Shapley value based on sampling",
Castro et al, 2009 (https://www.sciencedirect.com/science/article/pii/S0305054808000804)

samples : integer (default 5)
Defined the number of samples for each input feature.
Notice that evaluating a model samples * n_input_feature times might take a while.

sampling_dims : list of dimension indexes to run sampling on (feature dimensions).
By default, all dimensions except the batch dimension will be sampled.
For example, with a 4-D tensor that contains color images, single color channels are sampled.
To sample pixels, instead, use sampling_dims=[1,2]
"""


class ShapleySampling(PerturbationBasedMethod):

    def __init__(self, T, X, session, keras_learning_phase, feed_dict, samples=5, sampling_dims=None):
        super(ShapleySampling, self).__init__(T, X, session, keras_learning_phase, feed_dict)
        if self.has_multiple_inputs:
            raise RuntimeError('Multiple inputs not yet supported for perturbation methods')
        dims = len(X.shape)
        if sampling_dims is not None:
            if not 0 < len(sampling_dims) <= (dims - 1):
                raise RuntimeError('sampling_dims must be a list containing 1 to %d elements' % (dims-1))
            if 0 in sampling_dims:
                raise RuntimeError('Cannot sample batch dimension: remove 0 from sampling_dims')
            if any([x < 1 or x > dims-1 for x in sampling_dims]):
                raise RuntimeError('Invalid value in sampling_dims')
        else:
            sampling_dims = list(range(1, dims))

        self.samples = samples
        self.sampling_dims = sampling_dims

    def run(self, xs, ys=None, batch_size=None):
        xs_shape = list(xs.shape)
        batch_size = xs.shape[0]
        n_features = int(np.asscalar(np.prod([xs.shape[i] for i in self.sampling_dims])))
        result = np.zeros((xs_shape[0], n_features))

        run_shape = list(xs_shape)  # a copy
        run_shape = np.delete(run_shape, self.sampling_dims).tolist()
        run_shape.insert(1, -1)

        reconstruction_shape = [xs_shape[0]]
        for j in self.sampling_dims:
            reconstruction_shape.append(xs_shape[j])

        for r in range(self.samples):
            p = np.random.permutation(n_features)
            x = xs.copy().reshape(run_shape)
            y = None
            for i in p:
                if y is None:
                    y = self._session_run(self.T, x.reshape(xs_shape), ys, batch_size)
                x[:, i] = 0
                y0 = self._session_run(self.T, x.reshape(xs_shape), ys, batch_size)
                delta = y - y0
                delta_aggregated = np.sum(delta.reshape((batch_size, -1)), -1, keepdims=False)
                result[:, i] += delta_aggregated
                y = y0

        shapley = result / self.samples
        return shapley.reshape(reconstruction_shape)


# -----------------------------------------------------------------------------
# END ATTRIBUTION METHODS
# -----------------------------------------------------------------------------


attribution_methods = OrderedDict({
    'zero': (DummyZero, 0),
    'saliency': (Saliency, 1),
    'grad*input': (GradientXInput, 2),
    'intgrad': (IntegratedGradients, 3),
    'elrp': (EpsilonLRP, 4),
    'deeplift': (DeepLIFTRescale, 5),
    'occlusion': (Occlusion, 6),
    'shapley_sampling': (ShapleySampling, 7)
})



@ops.RegisterGradient("DeepExplainGrad")
def deepexplain_grad(op, grad):
    global _ENABLED_METHOD_CLASS, _GRAD_OVERRIDE_CHECKFLAG
    _GRAD_OVERRIDE_CHECKFLAG = 1
    if _ENABLED_METHOD_CLASS is not None \
            and issubclass(_ENABLED_METHOD_CLASS, GradientBasedMethod):
        return _ENABLED_METHOD_CLASS.nonlinearity_grad_override(op, grad)
    else:
        return original_grad(op, grad)


class DeepExplain(object):

    def __init__(self, graph=None, session=tf.get_default_session()):
        self.method = None
        self.batch_size = None
        self.session = session
        self.graph = session.graph if graph is None else graph
        self.graph_context = self.graph.as_default()
        self.override_context = self.graph.gradient_override_map(self.get_override_map())
        self.keras_phase_placeholder = None
        self.context_on = False
        if self.session is None:
            raise RuntimeError('DeepExplain: could not retrieve a session. Use DeepExplain(session=your_session).')

    def __enter__(self):
        # Override gradient of all ops created in context
        self.graph_context.__enter__()
        self.override_context.__enter__()
        self.context_on = True
        return self

    def __exit__(self, type, value, traceback):
        self.graph_context.__exit__(type, value, traceback)
        self.override_context.__exit__(type, value, traceback)
        self.context_on = False

    def get_explainer(self, method, T, X, **kwargs):
        if not self.context_on:
            raise RuntimeError('Explain can be called only within a DeepExplain context.')
        global _ENABLED_METHOD_CLASS, _GRAD_OVERRIDE_CHECKFLAG
        self.method = method
        if self.method in attribution_methods:
            method_class, method_flag = attribution_methods[self.method]
        else:
            raise RuntimeError('Method must be in %s' % list(attribution_methods.keys()))
        if isinstance(X, list):
            for x in X:
                if 'tensor' not in str(type(x)).lower():
                    raise RuntimeError('If a list, X must contain only Tensorflow Tensor objects')
        else:
            if 'tensor' not in str(type(X)).lower():
                raise RuntimeError('X must be a Tensorflow Tensor object or a list of them')

        if 'tensor' not in str(type(T)).lower():
            raise RuntimeError('T must be a Tensorflow Tensor object')

        logging.info('DeepExplain: running "%s" explanation method (%d)' % (self.method, method_flag))
        self._check_ops()
        _GRAD_OVERRIDE_CHECKFLAG = 0

        _ENABLED_METHOD_CLASS = method_class
        method = _ENABLED_METHOD_CLASS(T, X,
                                       self.session,
                                       keras_learning_phase=self.keras_phase_placeholder,
                                       **kwargs)

        if issubclass(_ENABLED_METHOD_CLASS, GradientBasedMethod) and _GRAD_OVERRIDE_CHECKFLAG == 0:
            warnings.warn('DeepExplain detected you are trying to use an attribution method that requires '
                          'gradient override but the original gradient was used instead. You might have forgot to '
                          '(re)create your graph within the DeepExlain context. Results are not reliable!')
        _ENABLED_METHOD_CLASS = None
        _GRAD_OVERRIDE_CHECKFLAG = 0
        self.keras_phase_placeholder = None
        return method

    def explain(self, method, T, X, xs, ys=None, batch_size=None, **kwargs):
        explainer = self.get_explainer(method, T, X, **kwargs)
        return explainer.run(xs, ys, batch_size)

    @staticmethod
    def get_override_map():
        return dict((a, 'DeepExplainGrad') for a in SUPPORTED_ACTIVATIONS)

    def _check_ops(self):
        """
        Heuristically check if any op is in the list of unsupported activation functions.
        This does not cover all cases where explanation methods would fail, and must be improved in the future.
        Also, check if the placeholder named 'keras_learning_phase' exists in the graph. This is used by Keras
         and needs to be passed in feed_dict.
        :return:
        """
        g = tf.get_default_graph()
        for op in g.get_operations():
            if len(op.inputs) > 0 and not op.name.startswith('gradients'):
                if op.type in UNSUPPORTED_ACTIVATIONS:
                    warnings.warn('Detected unsupported activation (%s). '
                                  'This might lead to unexpected or wrong results.' % op.type)
            elif 'keras_learning_phase' in op.name:
                self.keras_phase_placeholder = op.outputs[0]



from __future__ import (division, print_function, unicode_literals)

import numpy as np

from tensorflow_forward_ad import logger
from tensorflow_forward_ad.cbfs import bfs

log = logger.get()


def get_path_cover(adj, src, dst, bfs_cache=None):
  """Generates a list of nodes that contains all the nodes in between the
  source and the destination (endpoints included).

  Args:
    adj: Adjacency matrix.
    src: Source node index, 0-based.
    dst: Destination node index, 0-based.

  Returns:
    cover: List of node indices, in traversal order.
  """
  return get_path_cover_multi_src(adj, [src], dst, bfs_cache=bfs_cache)


def get_path_cover_multi_src(adj, src_list, dst, bfs_cache=None):
  """Generates a list of nodes that contains all the nodes in between
  multiple sources and a destination (endpoints included).

  Args:
    adj: Adjacency matrix.
    src: Source node index list, 0-based.
    dst: Destination node index, 0-based.

  Returns:
    cover: List of node indices, in traversal order.
  """
  num_nodes = adj.shape[0]
  if bfs_cache is None:
    bfs_cache = np.zeros([num_nodes, num_nodes], dtype=np.int8)
  cover = []
  for src in src_list:
    if src == dst:
      return cover
    if src > dst:
      raise Exception("Source must be smaller than destination.")
    if src >= num_nodes or dst >= num_nodes:
      raise Exception("Node index must be smaller than number of nodes.")
    if not bfs(adj, src, dst, cache=bfs_cache):
      log.warning("Source is not connected to destination.")
  cover.extend(src_list)
  start = min(src_list) + 1
  for idx in range(start, dst):
    # If the node is connected to both source and destination, add.
    if idx in src_list:
      continue
    connect_dst = bfs(adj, idx, dst, cache=bfs_cache)
    if not connect_dst:
      continue
    for src in src_list:
      connect_src = bfs(adj, src, idx, cache=bfs_cache)
      if connect_src:
        break
    if connect_src and connect_dst:
      cover.append(idx)
  cover.append(dst)
  return cover


def format_tensor_name(name):
  """Strips the tensor name to reflect the op name."""

  if name.startswith("^"):
    name_old = name
    name = name.strip("^")
    log.warning("Changing \"{}\" to \"{}\"".format(name_old, name))
  return name.split(":")[0]
  # return name


def convert_node_list_to_adj_mat(node_list):
  """Converts a node list into an adjacency matrix.

  Args:
    node_list: List of NodeDef objects.

  Returns:
    adj_mat: Adjacency matrix, np.array.
    ord_table: Mapping from node name to index, in a graph traversal order.
  """
  idx_table = {}
  ord_table = {}

  def add_node(table, node):
    if node.name in table:
      return
    for subnode_name in node.input:
      # Not sure if this will fix the variable
      # assign issue.
      # sbn = subnode_name.strip("^")\
      sbn = subnode_name
      sbn = format_tensor_name(sbn)
      if sbn not in table:
        if sbn not in idx_table:
          log.fatal("Not found {}".format(sbn))
          continue
        add_node(table, idx_table[sbn])
    table[node.name] = len(table)

  for node in node_list:
    idx_table[node.name] = node

  for node in node_list:
    add_node(ord_table, node)

  NN = len(node_list)
  adj_mat = [0] * (NN**2)
  for node in node_list:
    dst = ord_table[node.name]
    for inp in node.input:
      inp_ = format_tensor_name(inp)
      src = ord_table[inp_]
      adj_mat[src * NN + dst] = 1
  adj_mat = np.array(adj_mat, dtype=np.int8).reshape([NN, NN])
  return adj_mat, ord_table


def get_path_cover_str(node_list, src, dst):
  """Gets the path cover in string format.

  Args:
    node_list: List of NodeDef objects.
    src: Source node name, string.
    dst: Destination node name, string.

  Returns:
    list of node defs to traverse.
  """
  return get_path_cover_str_list(node_list, src, [dst])


def get_path_cover_str_list(node_list, src, dst_list):
  """Gets the path cover in string format.

  Args:
    node_list: List of NodeDef objects.
    src: Source node name, string.
    dst: Destination node name, list of string.

  Returns:
    list of node defs to traverse.
  """
  return get_path_cover_str_list_list(node_list, [src], dst_list)


def get_path_cover_str_list_list(node_list, src_list, dst_list):
  """Gets the path cover in string format.

  Args:
    node_list: List of NodeDef objects.
    src_list: Source node name, string, list of string.
    dst_list: Destination node name, list of string.

  Returns:
    list of node defs to traverse.
  """
  adj_mat, ord_table = convert_node_list_to_adj_mat(node_list)
  src_idx_list = [ord_table[format_tensor_name(src)] for src in src_list]
  dst_idx_list = [ord_table[format_tensor_name(dst)] for dst in dst_list]
  added_nodes = set()
  # Build inverse table from integer => string.
  ord_table_inv = [None] * len(node_list)
  num_nodes = len(node_list)
  bfs_cache = np.zeros([num_nodes, num_nodes], dtype=np.int8)
  for node, idx in ord_table.items():
    ord_table_inv[idx] = node
  path_cover_str = []
  # log.info("Computation graph BFS traversal...")
  # for dst_idx in tqdm(dst_idx_list):
  for dst_idx in dst_idx_list:
    path_cover = get_path_cover_multi_src(
        adj_mat, src_idx_list, dst_idx, bfs_cache=bfs_cache)
    for node_idx in path_cover:
      if node_idx not in added_nodes:
        path_cover_str.append(ord_table_inv[node_idx])
        added_nodes.add(node_idx)
  return path_cover_str



"""
Implements forward-mode automatic differentiation for TensorFlow.

Author: Mengye Ren (mren@cs.toronto.edu)

Usage:

Example 1:
  x = Variable(1.0)
  y = build_model(x)
  grad = forward_gradients(y, x)  # Gradient of y wrt. x, same shape as y.

Example 2::
  # When x is a vector, we compute the gradient of y wrt. a scalar v, with some
  # pre-specified value for the gradient of x wrt. v (by default a 1-vector).
  x = Variable([0.5, 0.8])
  y = build_model(x)
  grad = forward_gradients(y, x, grad_xs=tf.constant([1.0, 2.0]))

Example 3:
  x = [Variable(1.0), Variable(2.0)] # x can be a list of variables.
  y = build_model(x)
  # Grad is still the same dimension of y.
  grad = forward_gradients(y, x, grad_xs=[tf.constant(1.0), tf.constant(2.0)])

Example 4:
  x = Variable(1.0)
  y1 = build_model1(x)
  y2 = build_model2(x)
  # Grad is now a list: [dy1_dx, dy2_dx]
  grad = forward_gradients(y, x)
"""
from __future__ import (division, print_function, unicode_literals)

import numpy as np
import tensorflow as tf

from tensorflow.python.ops import gen_nn_ops
from tensorflow_forward_ad import logger
from tensorflow_forward_ad import maxpool_gradgrad
from tensorflow_forward_ad.graph import get_path_cover_str_list_list

log = logger.get()

###############################################################################
# Gradient operator registry utilities.
###############################################################################
FWGRAD_REGISTRY = {}
ELEM_REGISTRY = {}
NODEKW_REGISTRY = {}


def RegisterFwGrad(op, elemwise=False):
  """Registers a forward-mode gradient operator.

  Args:
    op: Name of the op to be differentiated.
    elemwise: Whether the gradient op is element-wise.

  Returns:
    decorator: A function that accepts a function (the gradient op) as input,
    that associates f with the original op.
  """

  def decorator(f):
    FWGRAD_REGISTRY[op] = f
    ELEM_REGISTRY[f] = elemwise

  return decorator


def RegisterNodeKeyword(op):
  """Registers a keyword extractor from NodeDef object.

  Args:
    op: Name of the op to be extracted.

  Returns:
    decorator: A function that accepts a function (the extractor function) as
    input that associates f  with the original op.
  """

  def decorator(f):
    NODEKW_REGISTRY[op] = f

  return decorator


def GetFwGrad(op):
  """Returns the forward mode gradient op that associates with the given op."""
  return FWGRAD_REGISTRY.get(op, None)


def GetNodeKeyword(op):
  """Returns the keyword extractor that associates with the given op."""
  return NODEKW_REGISTRY.get(op, None)


def GetElemwise(f):
  """Returns whether the gradient op elemwise with full forward gradients."""
  return ELEM_REGISTRY.get(f, False)


def NoFwGrad(op):

  def no_grad(*args, **kwargs):
    return None

  FWGRAD_REGISTRY[op] = no_grad


###############################################################################
# Dense operators. Not elemwise.
###############################################################################
@RegisterFwGrad("MatMul", elemwise=False)
def MatMul_FwGrad(op,
                  dx,
                  dw,
                  transpose_a=False,
                  transpose_b=False,
                  adjoint_a=False,
                  adjoint_b=False,
                  _op_table=None,
                  _grad_table=None):
  """Forward gradient operator for matrix multiplication.
  Note: only works when dim(dx) = x, i.e. the variable to be differentiated is
  a scalar.
  """
  x = op.inputs[0]
  w = op.inputs[1]
  if dx is not None:
    dy_x = tf.matmul(
        dx,
        w,
        transpose_a=transpose_a,
        transpose_b=transpose_b,
        adjoint_a=adjoint_a,
        adjoint_b=adjoint_b)
  if dw is not None:
    dy_w = tf.matmul(
        x,
        dw,
        transpose_a=transpose_a,
        transpose_b=transpose_b,
        adjoint_a=adjoint_a,
        adjoint_b=adjoint_b)
  if dx is None and dw is None:
    return None
  elif dx is not None and dw is None:
    return dy_x
  elif dw is not None and dx is None:
    return dy_w
  else:
    return tf.add(dy_x, dy_w)


@RegisterNodeKeyword("MatMul")
def MatMul_NodeKeyword(node):
  kwargs = {}
  kwdict = dict(node.attr)
  if "transpose_a" in kwdict:
    kwargs["transpose_a"] = kwdict["transpose_a"].b
  if "transpose_b" in kwdict:
    kwargs["transpose_b"] = kwdict["transpose_b"].b
  if "adjoint_a" in kwdict:
    kwargs["adjoint_a"] = kwdict["adjoint_a"].b
  if "adjoint_b" in kwdict:
    kwargs["adjoint_b"] = kwdict["adjoint_b"].b
  return kwargs


@RegisterFwGrad("Conv2D", elemwise=False)
def Conv2D_FwGrad(op,
                  dx,
                  dw,
                  strides=[1, 1, 1, 1],
                  padding="SAME",
                  data_format="NHWC",
                  use_cudnn_on_gpu=None,
                  _op_table=None,
                  _grad_table=None):
  """Forward gradient operator for 2D convolution.
  Note: only works when dim(dx) = x, i.e. the variable to be differentiated is
  a scalar.

  y = tf.conv2d(x, w, strides, padding)
  Now computes dy_dv given dx_dv and dw_dv.

  Args:
    x: Input tensor, 4D tensor, [N, H, W, C].
    w: Convolution filters, 4D tensor, [I, J, C, K].
    dx: Gradient of the input tensor, 4D tensor, [N, H, W, C].
    dw: Gradient of the convolution filters, 4D tensor, [I, J, C, K].
    strides: Strides, list of integers.
    padding: Padding, string, "SAME" or "VALID".
    data_format: "NHWC" or "NCHW".
    use_cudnn_on_gpu: Whether to use the cuDNN library on GPU.

  Returns:
    dy_dv: The gradient of y wrt. the variable to be differentiated.
  """
  x = op.inputs[0]
  w = op.inputs[1]
  if dx is not None:
    dy_x = tf.nn.conv2d(
        dx,
        w,
        strides,
        padding,
        data_format=data_format,
        use_cudnn_on_gpu=use_cudnn_on_gpu)
  if dw is not None:
    dy_w = tf.nn.conv2d(
        x,
        dw,
        strides,
        padding,
        data_format=data_format,
        use_cudnn_on_gpu=use_cudnn_on_gpu)
  if dx is None and dw is None:
    return None
  elif dx is not None and dw is None:
    return dy_x
  elif dw is not None and dx is None:
    return dy_w
  else:
    return tf.add(dy_x, dy_w)


@RegisterNodeKeyword("Conv2D")
def Conv2D_NodeKeyword(node):
  kwargs = {}
  kwdict = dict(node.attr)
  kwargs["strides"] = list(kwdict["strides"].list.i)
  kwargs["padding"] = kwdict["padding"].s
  kwargs["data_format"] = kwdict["data_format"].s
  kwargs["use_cudnn_on_gpu"] = kwdict["use_cudnn_on_gpu"].b
  return kwargs


@RegisterFwGrad("Conv2DBackpropFilter", elemwise=False)
def Conv2DBackpropFilter_FwGrad(op,
                                dx,
                                dy,
                                dz,
                                strides=[1, 1, 1, 1],
                                padding="SAME",
                                data_format="NHWC",
                                use_cudnn_on_gpu=None,
                                _op_table=None,
                                _grad_table=None):
  """Forward gradient operator of conv 2d gradient wrt. filters.
  Args:
    op: Conv2DBackpropFilter operator.
    dx: Forward gradient of the input to Conv2D.
    dy: Forward gradient of the filter shape.
    dz: Forward gradient of the backward gradient of the output of the Conv2D.
  """
  # dx = dx_dr
  # dz = d(dE_dy)_dr
  dE_dy = op.inputs[2]  # dE_dy
  x = op.inputs[0]
  fshape = op.inputs[1]
  if dx is not None:
    dy_x = tf.nn.conv2d_backprop_filter(
        dx,
        fshape,
        dE_dy,
        strides,
        padding,
        data_format=data_format,
        use_cudnn_on_gpu=use_cudnn_on_gpu)
  if dz is not None:
    dy_z = tf.nn.conv2d_backprop_filter(
        x,
        fshape,
        dz,
        strides,
        padding,
        data_format=data_format,
        use_cudnn_on_gpu=use_cudnn_on_gpu)
  if dx is None and dz is None:
    return None
  elif dx is not None and dz is None:
    return dy_x
  elif dz is not None and dx is None:
    return dy_z
  else:
    return tf.add(dy_x, dy_z)


@RegisterNodeKeyword("Conv2DBackpropFilter")
def Conv2DBackpropFilter_NodeKeyword(node):
  kwargs = {}
  kwdict = dict(node.attr)
  kwargs["strides"] = list(kwdict["strides"].list.i)
  kwargs["padding"] = kwdict["padding"].s
  kwargs["data_format"] = kwdict["data_format"].s
  kwargs["use_cudnn_on_gpu"] = kwdict["use_cudnn_on_gpu"].b
  return kwargs


@RegisterFwGrad("Conv2DBackpropInput", elemwise=False)
def Conv2DBackpropInput_FwGrad(op,
                               dx,
                               dy,
                               dz,
                               strides=[1, 1, 1, 1],
                               padding="SAME",
                               data_format="NHWC",
                               use_cudnn_on_gpu=None,
                               _op_table=None,
                               _grad_table=None):
  """Forward gradient operator of conv 2d gradient wrt. input.
  Args:
    op: Conv2DBackpropInput operator.
    dx: Forward gradient of the input shape.
    dy: Forward gradient of the filter.
    dz: Forward gradient of the backward gradient of the output of the Conv2D.
  """
  # dz = d(dE_dy)_dr
  dE_dy = op.inputs[2]  # dE_dy
  f = op.inputs[1]
  df = dy
  xshape = op.inputs[0]
  if df is not None:
    dy_f = tf.nn.conv2d_backprop_input(
        xshape,
        df,
        dE_dy,
        strides,
        padding,
        data_format=data_format,
        use_cudnn_on_gpu=use_cudnn_on_gpu)
  if dz is not None:
    dy_z = tf.nn.conv2d_backprop_input(
        xshape,
        f,
        dz,
        strides,
        padding,
        data_format=data_format,
        use_cudnn_on_gpu=use_cudnn_on_gpu)
  if df is None and dz is None:
    return None
  elif df is not None and dz is None:
    return dy_f
  elif dz is not None and df is None:
    return dy_z
  else:
    return tf.add(dy_f, dy_z)


@RegisterNodeKeyword("Conv2DBackpropInput")
def Conv2DBackpropInput_NodeKeyword(node):
  kwargs = {}
  kwdict = dict(node.attr)
  kwargs["strides"] = list(kwdict["strides"].list.i)
  kwargs["padding"] = kwdict["padding"].s
  kwargs["data_format"] = kwdict["data_format"].s
  kwargs["use_cudnn_on_gpu"] = kwdict["use_cudnn_on_gpu"].b
  return kwargs


@RegisterFwGrad("MaxPool", elemwise=False)
def MaxPool_FwGrad(op,
                   dx,
                   ksize=[1, 2, 2, 1],
                   strides=[1, 2, 2, 1],
                   padding="SAME",
                   _op_table=None,
                   _grad_table=None):
  """Forward gradient operator for max pooling.

  Args:
    x: Input tensor, 4D tensor, [N, H, W, C].
    dx: Gradient of the input tensor, 4D tensor, [N, H, W, C].
    ksize: Kernel size of the max pooling operator, list of integers.
    strides: Strides of the max pooling operator, list of integers.
    padding: Padding, string, "SAME" or "VALID".
    data_format: "NHWC" or "NCHW".
  """
  if dx is None:
    return None
  x = op.inputs[0]
  y = op.outputs[0]
  _, argmax = tf.nn.max_pool_with_argmax(x, ksize, strides, padding)
  dx_flat = tf.reshape(dx, [-1])
  argmax_flat = tf.reshape(argmax, [-1])
  y_zero = tf.zeros_like(y, dtype=argmax.dtype)
  x_shape = tf.cast(tf.shape(x), argmax.dtype)
  batch_dim = tf.reshape(
      tf.range(x_shape[0], dtype=argmax.dtype), [-1, 1, 1, 1])
  nelem = tf.reduce_prod(x_shape[1:])
  batch_dim *= nelem
  batch_dim += y_zero
  batch_dim = tf.reshape(batch_dim, [-1])
  argmax_flat += batch_dim
  dx_sel = tf.gather(dx_flat, argmax_flat)
  dy = tf.reshape(dx_sel, tf.shape(argmax))
  return dy


@RegisterNodeKeyword("MaxPool")
def MaxPool_NodeKeyword(node):
  kwargs = {}
  kwdict = dict(node.attr)
  kwargs["strides"] = list(kwdict["strides"].list.i)
  kwargs["ksize"] = list(kwdict["ksize"].list.i)
  kwargs["padding"] = kwdict["padding"].s
  return kwargs


@RegisterFwGrad("MaxPoolGrad", elemwise=False)
def MaxPoolGrad_FwGrad(op,
                       dx,
                       dy,
                       dz,
                       ksize=[1, 2, 2, 1],
                       strides=[1, 2, 2, 1],
                       data_format="NHWC",
                       padding="SAME",
                       _op_table=None,
                       _grad_table=None):
  """Forward gradient operator for the backward gradient of max pooling.

  Args:
    op: MaxPoolGrad operator.
    dx: Forward gradient to the input of MaxPool.
    dy: Forward gradient to the output of MaxPool.
    dz: Forward gradient to the backward gradient of the output of MaxPool.
  """
  if dz is None:
    return None
  return gen_nn_ops._max_pool_grad(
      op.inputs[0],
      op.inputs[1],
      dz,
      ksize,
      strides,
      padding,
      data_format=data_format)


@RegisterNodeKeyword("MaxPoolGrad")
def MaxPoolGrad_NodeKeyword(node):
  kwargs = {}
  kwdict = dict(node.attr)
  kwargs["strides"] = list(kwdict["strides"].list.i)
  kwargs["ksize"] = list(kwdict["ksize"].list.i)
  kwargs["padding"] = kwdict["padding"].s
  return kwargs


@RegisterFwGrad("StridedSlice", elemwise=False)
def StridedSlice_FwGrad(op, dx, dy, dz, du, _op_table=None, _grad_table=None):
  if dx is None:
    return None
  y = op.inputs[1]
  z = op.inputs[2]
  u = op.inputs[3]
  return tf.strided_slice(dx, begin=y, end=z, strides=u)


###############################################################################
# Element-wise operators. elemwise.
###############################################################################
@RegisterFwGrad("Relu", elemwise=True)
def Relu_FwGrad(op, dx, _op_table=None, _grad_table=None):
  if dx is None:
    return None
  x = op.inputs[0]
  return tf.multiply(dx, tf.cast(tf.greater(x, 0.0), x.dtype))


@RegisterFwGrad("ReluGrad", elemwise=True)
def ReluGrad_FwGrad(op, dx, dy, _op_table=None, _grad_table=None):
  """Forward gradient operator for the gradient of ReLU.
  f = ReLU(v)
  forward pass: f(u)
  dx = d(dE_df)_dr
  dy = df_dr
  Want d(dE_dv)_dr

  Args:
    op: ReluGrad op.
    dx: Forward gradient of the backward gradient of the ReLU output.
    dy: Forward gradient of ReLU output.
  """
  if dx is None:
    return None
  x = op.inputs[0]  # Backward gradient of the ReLU output, dE_df.
  y = op.inputs[1]  # ReLU output, f.
  return tf.multiply(dx, tf.cast(tf.greater(y, 0.0), dx.dtype))


@RegisterFwGrad("Add", elemwise=True)
def Add_FwGrad(op, dx, dy, _op_table=None, _grad_table=None):
  if dx is None and dy is None:
    return None
  else:
    if dx is None:
      x = op.inputs[0]
      dx = tf.zeros_like(x)
    if dy is None:
      y = op.inputs[1]
      dy = tf.zeros_like(y)
    return tf.add(dx, dy)


@RegisterFwGrad("AddV2", elemwise=True)
def Add_FwGrad(op, dx, dy, _op_table=None, _grad_table=None):
  if dx is None and dy is None:
    return None
  else:
    if dx is None:
      x = op.inputs[0]
      dx = tf.zeros_like(x)
    if dy is None:
      y = op.inputs[1]
      dy = tf.zeros_like(y)
    return tf.add(dx, dy)


@RegisterFwGrad("BiasAdd", elemwise=True)
def Add_FwGrad(op, dx, dy, _op_table=None, _grad_table=None):
  if dx is None and dy is None:
    return None
  else:
    if dx is None:
      x = op.inputs[0]
      dx = tf.zeros_like(x)
    if dy is None:
      y = op.inputs[1]
      dy = tf.zeros_like(y)
    return tf.add(dx, dy)



@RegisterFwGrad("Sub", elemwise=True)
def Sub_FwGrad(op, dx, dy, _op_table=None, _grad_table=None):
  if dx is None and dy is None:
    return None
  else:
    if dx is None:
      x = op.inputs[0]
      dx = tf.zeros_like(x)
    if dy is None:
      y = op.inputs[1]
      dy = tf.zeros_like(y)
    return tf.subtract(dx, dy)


@RegisterFwGrad("Neg", elemwise=True)
def Neg_FwGrad(op, dx, _op_table=None, _grad_table=None):
  if dx is None:
    return None
  return tf.negative(dx)


@RegisterFwGrad("AddN", elemwise=True)
def AddN_FwGrad(*args, **kwargs):
  no_none = list(filter(lambda x: x is not None, args[1:]))
  if len(no_none) > 0:
    return tf.add_n(no_none)
  else:
    return None


@RegisterFwGrad("Mul", elemwise=True)
def Mul_FwGrad(op, dx, dy, _op_table=None, _grad_table=None):
  x = op.inputs[0]
  y = op.inputs[1]
  if dx is None and dy is None:
    return None
  elif dx is not None and dy is None:
    return tf.multiply(y, dx)
  elif dy is not None and dx is None:
    return tf.multiply(x, dy)
  else:
    return tf.add(tf.multiply(x, dy), tf.multiply(y, dx))


if tf.__version__.startswith("0"):
  divname = "Div"
else:
  divname = "RealDiv"


@RegisterFwGrad(divname, elemwise=True)
def Div_FwGrad(op, dx, dy, _op_table=None, _grad_table=None):
  x = op.inputs[0]
  y = op.inputs[1]
  if dx is None and dy is None:
    return None
  elif dx is not None and dy is None:
    return tf.divide(dx, y)
  elif dy is not None and dx is None:
    return -tf.divide(x * dy, y**2)
  else:
    return tf.divide(y * dx - x * dy, y**2)


@RegisterFwGrad("Reciprocal", elemwise=True)
def Reciprocal_FwGrad(op, dx, _op_table=None, _grad_table=None):
  x = op.inputs[0]
  if dx is None:
    return None
  return -1 / tf.square(x) * dx


@RegisterFwGrad("Square", elemwise=True)
def Square_FwGrad(op, dx, _op_table=None, _grad_table=None):
  x = op.inputs[0]
  if dx is None:
    return None
  return tf.multiply(tf.multiply(tf.constant(2.0, x.dtype), x), dx)


@RegisterFwGrad("Sqrt", elemwise=True)
def Sqrt_FwGrad(op, dx, _op_table=None, _grad_table=None):
  y = op.outputs[0]
  if dx is None:
    return None
  return tf.div(0.5 * dx, y + tf.cast(tf.equal(y, 0.0), y.dtype)) * tf.cast(
      tf.greater(y, 0.0), y.dtype)


@RegisterFwGrad("Pow", elemwise=True)
def Pow_FwGrad(op, dx, dy, _op_table=None, _grad_table=None):
  x = op.inputs[0]
  y = op.inputs[1]
  if dx is None and dy is None:
    return None
  elif dx is not None and dy is None:
    return y * x**(y - 1) * dx
  elif dy is not None and dx is None:
    return x**y * dy * tf.log(x)
  else:
    return y * x**(y - 1) * dx + x**y * dy * tf.log(x)


@RegisterFwGrad("Identity", elemwise=True)
def Identity_FwGrad(op, dx, _op_table=None, _grad_table=None):
  return dx


@RegisterFwGrad("PreventGradient", elemwise=True)
def Identity_FwGrad(op, dx, _op_table=None, _grad_table=None):
  return dx


@RegisterFwGrad("Print", elemwise=True)
def Print_FwGrad(*args, **kwargs):
  dx = args[1]
  return dx


@RegisterFwGrad("Sigmoid", elemwise=True)
def Sigmoid_FwGrad(op, dx, _op_table=None, _grad_table=None):
  if dx is None:
    return None
  y = op.outputs[0]
  return tf.multiply(tf.multiply(dx, (1 - y)), y)


@RegisterFwGrad("SigmoidGrad", elemwise=True)
def SigmoidGrad_FwGrad(op, dx, dy, _op_table=None, _grad_table=None):
  """Forward gradient of SigmoidGrad.
  f = Sigmoid(v).
  dx = df_dr
  dy = d(dE_df)_dr
  Want: d(dE_dv)_dr

  Args:
    op: SigmoidGrad op.
    dx: Forward gradient of the sigmoid.
    dy: Forward gradient of the backward gradient of the sigmoid output.
  """
  f = op.inputs[0]
  f_op = _op_table[f.op.name]
  v = f_op.inputs[0]
  dv_dr = _grad_table[v.name]
  df_dv = f * (1 - f)
  d2E_dfdr = dy

  if dv_dr is None:
    if d2E_dfdr is None:
      return None
    else:
      return d2E_dfdr * df_dv
  else:
    dE_df = op.inputs[1]
    d2f_dv2 = df_dv * (1 - 2 * f)
    df_dr = dx
    d2f_dvdr = df_dr * (1 - 2 * f)
    if d2E_dfdr is None:
      return d2f_dvdr * dE_df
    else:
      return d2E_dfdr * df_dv + d2f_dvdr * dE_df


@RegisterFwGrad("Tanh", elemwise=True)
def Tanh_FwGrad(op, dx, _op_table=None, _grad_table=None):
  y = op.outputs[0]
  if dx is None:
    return None
  return tf.multiply(dx, 1 - y**2)


@RegisterFwGrad("TanhGrad", elemwise=True)
def TanhGrad_FwGrad(op, dx, dy, _op_table=None, _grad_table=None):
  """Forward gradient of SigmoidGrad.
  f = Sigmoid(v).
  dx = df_dr
  dy = d(dE_df)_dr
  Want: d(dE_dv)_dr

  Args:
    op: TanhGrad op.
    dx: Forward gradient of the sigmoid.
    dy: Forward gradient of the backward gradient of the sigmoid output.
  """
  f = op.inputs[0]
  f_op = _op_table[f.op.name]
  v = f_op.inputs[0]
  dv_dr = _grad_table[v.name]
  df_dv = 1 - tf.square(f)
  d2E_dfdr = dy

  if dv_dr is None:
    if d2E_dfdr is None:
      return None
    else:
      return dy * df_dv
  else:
    dE_df = op.inputs[1]
    d2f_dv_dr = -2 * f * dx
    if d2E_dfdr is None:
      return d2f_dv_dr * dE_df
    else:
      return d2E_dfdr * df_dv + d2f_dv_dr * dE_df


@RegisterFwGrad("Softmax", elemwise=True)
def Softmax_FwGrad(op, dx, _op_table=None, _grad_table=None):
  """Forward gradient operator for softmax."""
  y = op.outputs[0]
  if dx is None:
    return None
  return tf.subtract(
      tf.multiply(y, dx),
      tf.multiply(y, tf.reduce_sum(tf.multiply(dx, y), [1], keep_dims=True)))


@RegisterFwGrad("Log", elemwise=True)
def Log_FwGrad(op, dx, _op_table=None, _grad_table=None):
  x = op.inputs[0]
  if dx is None:
    return None
  return 1 / x * dx


@RegisterFwGrad("SparseSoftmaxCrossEntropyWithLogits", elemwise=True)
def SparseSoftmaxCrossEntropyWithLogits_FwGrad(op,
                                               dx,
                                               dy,
                                               _op_table=None,
                                               _grad_table=None):
  """Forward gradient operator of sparse softmax cross entropy."""
  grad = op.outputs[1]  # This is already computed in the forward pass.
  x = op.inputs[0]
  if dx is None:
    return None
  y = tf.nn.softmax(x)
  grad_grad = tf.subtract(
      tf.multiply(y, dx),
      tf.multiply(y, tf.reduce_sum(tf.multiply(dx, y), [1], keep_dims=True)))
  return tf.reduce_sum(tf.multiply(grad, dx), [1]), grad_grad


@RegisterFwGrad("Transpose", elemwise=True)
def Transpose_FwGrad(op, dx, dy, _op_table=None, _grad_table=None):
  y = op.inputs[1]
  return tf.transpose(dx, y)


###############################################################################
# Shape sensitive operators. Not elemwise.
###############################################################################
@RegisterFwGrad("Pack", elemwise=True)
def Pack_FwGrad(*args, **kwargs):
  dx = args[1:]
  axis = kwargs["axis"]
  if all(map(lambda x: x is None, dx)):
    log.error("hey")
    return None
  else:
    ### Here we need to fill in zeros.
    def _mapper(_):
      dx = _[0]
      x = _[1]
      return dx if dx is not None else tf.zeros_like(x)

    dx = list(map(_mapper, zip(dx, list(args[0].inputs))))
    if tf.__version__.startswith("0"):
      return tf.pack(dx, axis=axis)
    else:
      return tf.stack(dx, axis=axis)


@RegisterNodeKeyword("Pack")
def PackGrad_NodeKeyword(node):
  kwargs = {}
  kwdict = dict(node.attr)
  kwargs["axis"] = kwdict["axis"].i
  return kwargs


@RegisterFwGrad("Reshape", elemwise=False)
def Reshape_FwGrad(op, dx, dy, _op_table=None, _grad_table=None):
  shape = op.inputs[1]
  if dx is None:
    return None
  return tf.reshape(dx, shape)


@RegisterFwGrad("Sum", elemwise=False)
def ReduceSum_FwGrad(op,
                     dx,
                     dy,
                     keep_dims=False,
                     _op_table=None,
                     _grad_table=None):
  ridx = op.inputs[1]
  if dx is None:
    return None
  return tf.reduce_sum(dx, keep_dims=keep_dims, reduction_indices=ridx)


@RegisterNodeKeyword("Sum")
def ReduceSum_NodeKeyword(node):
  kwargs = {}
  kwdict = dict(node.attr)
  kwargs["keep_dims"] = kwdict["keep_dims"].b
  return kwargs


@RegisterFwGrad("Mean", elemwise=False)
def ReduceMean_FwGrad(op,
                      dx,
                      dy,
                      keep_dims=False,
                      _op_table=None,
                      _grad_table=None):
  ridx = op.inputs[1]
  if dx is None:
    return None
  return tf.reduce_mean(dx, keep_dims=keep_dims, reduction_indices=ridx)


@RegisterNodeKeyword("Mean")
def ReduceMean_NodeKeyword(node):
  kwargs = {}
  kwdict = dict(node.attr)
  kwargs["keep_dims"] = kwdict["keep_dims"].b
  return kwargs


@RegisterFwGrad("Tile", elemwise=False)
def Tile_FwGrad(op, dx, dy, _op_table=None, _grad_table=None):
  y = op.inputs[1]
  if dx is None:
    return None
  return tf.tile(dx, y)


@RegisterFwGrad("ExpandDims", elemwise=False)
def ExpandDims_FwGrad(op, dx, dy, _op_table=None, _grad_table=None):
  y = op.inputs[1]
  if dx is None:
    return None
  return tf.expand_dims(dx, y)


###############################################################################
# No gradient operator defined.
###############################################################################
NoFwGrad("NoOp")
NoFwGrad("BroadcastGradientArgs")
NoFwGrad("Const")
NoFwGrad("Shape")
NoFwGrad("Rank")
NoFwGrad("Range")
NoFwGrad("InvertPermutation")


###############################################################################
# Utilities
###############################################################################
def value_or_list(v):
  if type(v) == list:
    return v
  else:
    return [v]


def get_path_cover(ys, xs, node_list):
  ys = value_or_list(ys)
  xs = value_or_list(xs)
  y_name_list = [yy.name for yy in ys]
  x_name_list = [xx.name for xx in xs]
  path_cover_str = get_path_cover_str_list_list(node_list, x_name_list,
                                                y_name_list)
  return path_cover_str


def get_node_kwargs(node):
  """Get additional attributes from the node as keyword arguments to the
  forward gradient operator.

  Args:
    node: NodeDef object.

  Returns:
    kwargs: Keyword arguments dictionary.
  """
  kwargs_fn = GetNodeKeyword(node.op)
  if kwargs_fn is None:
    return {}
  else:
    return kwargs_fn(node)


def get_node_arg_list(op, grad_table):
  # Builds an argument list.
  arg_list = []
  arg_list.append(op)
  for inp in op.inputs:
    if grad_table[inp.name] is None:
      log.warning("Node \"{}\" gradient does not exists.".format(inp.name))
      grad_table[inp.name] = None
    arg_list.append(grad_table[inp.name])
  return arg_list


def store_results(op, grads, grad_table):
  """Stores gradient results to gradient table."""
  if len(op.outputs) == 1:
    grad_table[op.outputs[0].name] = grads
  elif len(op.outputs) > 1:
    if grads is not None:
      for output, grad in zip(op.outputs, grads):
        grad_table[output.name] = grad
    else:
      for output in op.outputs:
        grad_table[output.name] = None


###############################################################################
# Main forward gradient function. Entry point.
###############################################################################
def forward_gradients_v2(ys, xs, grad_xs=None, gate_gradients=False):
  """Forward-mode pushforward analogous to the pullback defined by tf.gradients.
  With tf.gradients, grad_ys is the vector being pulled back, and here d_xs is
  the vector being pushed forward."""
  if type(ys) == list:
    v = [tf.ones_like(yy) for yy in ys]
  else:
    v = tf.ones_like(ys)  # dummy variable
  g = tf.gradients(ys, xs, grad_ys=v)
  return tf.gradients(g, v, grad_ys=grad_xs)


def forward_gradients(ys,
                      xs,
                      grad_xs=None,
                      gate_gradients=False,
                      name="gradients"):
  """Forward-mode automatic differentiation.
  Currently support single variable xs and single/multi variable(s) output ys.

  Args:
    ys: Output variable or a list of output variables.
    xs: Input variable or a list of input variables, v if grad_xs is not
    defined.
    grad_xs: Gradients of x to v, default is the one-vector [1, 1, ...]. This
    is similar to grad_ys in tf.gradients.
    gate_gadients: If True, add a tuple around the results.
    name: Name of the op.

  Returns:
    grad: Gradient of y wrt. v.
  """
  graph = tf.get_default_graph()
  graph_def = graph.as_graph_def()
  node_list = graph_def.node
  with tf.variable_scope(name):
    xs = value_or_list(xs)
    ys = value_or_list(ys)
    grad_xs = value_or_list(grad_xs)
    assert len(grad_xs) == len(xs), "Length of grad_xs and xs must equal."

    # Build path cover.
    path_cover_str = get_path_cover(ys, xs, node_list)

    # Initialize node table.
    # From node name to its node def object.
    node_table = dict(map(lambda x: (x.name, x), node_list))

    # Initialize op table.
    # From op name to operation object.
    op_table = dict(map(lambda x: (x.name, x), graph.get_operations()))

    # Initialize tensor table and gradient table.
    tensor_table = dict()  # From tensor name to its tensor object.
    grad_table = dict()  # From tensor name to its forward gradient.
    for node in node_list:
      for output in op_table[node.name].outputs:
        tensor_table[output.name] = output
        grad_table[output.name] = None

    # Set up initial condition.
    inp_names = set(map(lambda x: x.name.split(":")[0], xs))
    for ii, xx in enumerate(xs):
      if grad_xs is None or grad_xs[ii] is None:
        grad_xx = tf.ones_like(xx)
      else:
        grad_xx = grad_xs[ii]
      grad_table[xx.name] = grad_xx

    # Build gradient graph.
    for node_str in path_cover_str:
      if node_str in inp_names:
        continue
      node = node_table[node_str]

      # Build argument list.
      op = op_table[node.name]
      arg_list = get_node_arg_list(op, grad_table)
      kwargs = get_node_kwargs(node)
      kwargs["_op_table"] = op_table
      kwargs["_grad_table"] = grad_table

      # Call forward gradient function.
      fw_grad_fn = GetFwGrad(node.op)
      if fw_grad_fn is None:
        # raise Exception("Op \"{}\" has no FwGrad registered.".format(node.op))
        log.error("Op \"{}\" has no FwGrad registered.".format(node.op))
        node_grads = None
      else:
        # Compute forward gradient.
        node_grads = fw_grad_fn(*arg_list, **kwargs)

      # Store results.
      store_results(op, node_grads, grad_table)

    grads = list(map(lambda x: grad_table[x.name], ys))
    if gate_gradients:
      grads = tf.tuple(grads)
    return grads



import os
import time
import logging

# Tensorflow
import tensorflow as tf
import tensorflow_probability as tfp

from .util import scope_name

def layer(h, dim = None, fun = None, name = 'dense', l2_scale = 1e-50, l1_scale = 1e-50,
          drop = None,  batch_norm = None, custom = False, W = None, b = None, probability = False,
          kernel_initializer = tf.contrib.layers.xavier_initializer(uniform = False),
          bias_initializer = tf.zeros_initializer(), use_bias = True, graph = None, axis_batch = -2, match_dim = False, reuse = True):
    """ defines a layer using tf.layers """

    variable_scope = 'layer'

    if not reuse:
        variable_scope = scope_name(variable_scope)

    if l2_scale >= l1_scale:
        regularizer = tf.contrib.layers.l2_regularizer(scale = l2_scale)
    else:
        regularizer = tf.contrib.layers.l1_regularizer(scale = l1_scale)

    with tf.variable_scope(variable_scope, reuse=tf.AUTO_REUSE):

        if batch_norm is not None:
            h = tf.layers.batch_normalization(h, training=batch_norm)

        if not custom:
            print ('not custom')

            if W is not None:
                raise Exception('Warning: W is not being used with tf.layer.dense, set custom to True')

            if len(h.shape) != 2 and match_dim:
                raise Exception('Tensor with more than 2 dimensions is being used on tf.layer.dense')

            h = tf.layers.dense(inputs = h, units = dim, activation = fun,
                                kernel_initializer = kernel_initializer,
                                kernel_regularizer = regularizer,
                                use_bias = use_bias,
                                bias_initializer = bias_initializer,
                                name = name)

            with tf.variable_scope(name, reuse=tf.AUTO_REUSE):

                W = tf.get_variable('kernel')

                if use_bias:
                    if b is not None:
                        raise Exception('Warning: b is not being used with tf.layer.dense, set custom to True')
                    b = tf.get_variable('bias')

        else:
            logging.info('custom')
            if W is None or (b is None and use_bias):
                if match_dim:
                    shape = [str(d) for d in h.shape]
                    shape.pop(axis_batch)
                    shape.append(dim)
                    shape = [int(d) for d in shape]
                    shape_W = list(shape)
                    shape_b = [shape[0],shape[-1]]
                else:
                    shape_W = [int(h.shape[-1]),dim]
                    shape_b = [dim]

            kwargs = {'dtype'      : tf.float32,
                      'trainable'  : True
            }

            variable_scope = scope_name('dense')

            with tf.variable_scope('dense'):

                if W is None:

                    W = tf.get_variable(name = 'kernel',
                                        shape = shape_W,
                                        initializer = kernel_initializer,
                                        regularizer = regularizer,
                                        **kwargs)

                if use_bias:

                    if b is None:

                        b = tf.get_variable(name = 'bias',
                                            shape = shape_b,
                                            initializer = bias_initializer,
                                            regularizer = regularizer,
                                            **kwargs)
            if axis_batch == 0:
                h = tf.transpose(h,[1,0,2])
                h = tf.matmul(h,W)
                h = tf.transpose(h,[1,0,2])
            else:
                h = tf.matmul(h,W)
            if use_bias:
                h = h + b

            if fun is not None:
                h = fun(h)

        if drop is not None:
            h = tf.nn.dropout(h, drop)

    return(h, W, b)


def KL_divergence_normal(dist1, dist2 = None):
    """
    KL Divergence
    D_KL[N(mu_1,exp(logvar_1)I) || N(mu_2,exp(logvar_2)I)]
    """

    mu_1, var_1, logvar_1 = dist1

    if dist2 is None:

        loss = 0.5 * (mu_1**2 + var_1 - 1 - logvar_1)

    else:
        mu_2, var_2, logvar_2 = dist2
        loss = 0.5 * ((logvar2 - logvar_1)
                      - 1
                      + var_1 / var_2
                      + (mu_2 - mu_1)**2 / var_2
                     )

    loss = tf.identity(loss, "KL_Divergence_Normal")

    return loss


def create_variance(h, var_dependency, var_type, dim, l2_scale, l1_scale, drop,
                    batch_norm, custom, activation_fun = tf.nn.softplus,
                    kernel_initializer = None, bias_initializer = None, initializer = None, epsilon = 1e-8):
    """
    Input
        h: hidden layer
        var_dependency: True if varince is function of h
        var_type: diagonal, scalar, identity
        activation_fun: activation fun for the full connected layer predicting variance
        kernel_initializer: None if var_dependency is False
        bias_initializer: None if var_dependency is False
        initializer: None if var_dependency is True
    Return
        tensor
    """

    name = 'var'

    if var_dependency:

        if var_type == 'fullcov':

            h_var_diag,_,_ = layer(h, dim, fun = activation_fun, name = "{}_{}".format(name,'covar_vector'), l2_scale = l2_scale, l1_scale = l1_scale,
                      drop = drop,  batch_norm = batch_norm, custom = custom, kernel_initializer = kernel_initializer, bias_initializer = bias_initializer)

            h_var,_,_ = layer(h, dim * dim, fun = None, name = "{}_{}".format(name,'covar_vector'), l2_scale = l2_scale, l1_scale = l1_scale,
                      drop = drop,  batch_norm = batch_norm, custom = custom, kernel_initializer = kernel_initializer, bias_initializer = bias_initializer)

            h_var = tf.reshape(h_var, tf.shape(h)[:-1] + [dim, dim],
                               name = "{}_{}".format(name,'CovarianceMatrix'))

            h_var = tf.linalg.set_diag(h_var,h_var_diag,
                                       name = "{}_{}".format(name,'CovarianceMatrix'))

        elif var_type == 'identity':
                h_var = tf.ones(dim, name = 'var_diagonal')


        elif var_type == 'scalar':
            h_var_scalar,_,_ = layer(h, 1, fun = activation_fun, name = "{}_{}".format(name,'scalar'), l2_scale = l2_scale, l1_scale = l1_scale,
                      drop = drop,  batch_norm = batch_norm, custom = custom, kernel_initializer = kernel_initializer, bias_initializer = bias_initializer)
            h_var = tf.multiply(h_var_scalar, tf.ones(dim), name = 'diagonal')


        elif var_type == 'diagonal':
            h_var,_,_ = layer(h, dim, fun = activation_fun, name = "{}_{}".format(name,'diagonal'), l2_scale = l2_scale, l1_scale = l1_scale,
                      drop = drop,  batch_norm = batch_norm, custom = custom, kernel_initializer = kernel_initializer, bias_initializer = bias_initializer)


    else:

        # Set up variance first
        if var_type == 'identity':
                h_var = tf.ones(dim, name = 'var_diagonal')

        elif var_type == 'scalar':
                h_var_scalar = tf.get_variable("var_scalar", shape=[],
                                              initializer=initializer)
                h_var = tf.ones(dim) * h_var_scalar

        elif var_type == 'diagonal':
                h_var = tf.get_variable("var_diagonal", shape=[dim],
                                              initializer=initializer)

    return tf.add(h_var, epsilon)


class Distribution():

    """ Create probability layer """

    def __init__(self, h, dim, h_mu = None, h_std = None, fun = None, name = None, l2_scale = 1e-50, l1_scale = 1e-50,
                     drop = None,  batch_norm = None, custom = False,
                     epsilon = 1e-5, var_dependency = True, var_type = 'diagonal', activation_fun = tf.nn.softplus):

        if not var_type in ['scalar','diagonal','deterministic','identity']:
            raise Exception('Invalid input for var_type({})'.format(var_type))

        if h_mu is None:
            with tf.variable_scope('mu'):
                    h_mu,_,_ = layer(h, dim, fun = fun, name = "{}_{}".format(name,'mu'),
                                     l2_scale = l2_scale, l1_scale = l1_scale,
                                     drop = drop,  batch_norm = batch_norm, custom = custom)

        if var_type == 'deterministic':

            h_sample = h_mu
            h_std = None
            h_var = None
            h_logvar = None
            h_epsilon = None
            h_dist = None

        else:
            with tf.variable_scope('var'):

                if h_std is None:

                    h_std = create_variance(h = h, dim = dim, var_dependency = var_dependency, var_type = var_type, activation_fun = activation_fun,
                                            batch_norm = None, custom = custom, l2_scale = 1e-50, l1_scale = 1e-50, drop = None,
                                            kernel_initializer = tf.zeros_initializer,
                                            bias_initializer = tf.zeros_initializer,
                                            epsilon = epsilon)

                h_var = tf.square(h_std)
                h_logvar = tf.log(h_var + epsilon)

            if var_type == 'fullcov':
                h_dist = tfp.distributions.MultivariateNormalFullCovariance(loc = h_mu,
                                                                            covariance_matrix = h_std,
                                                                            allow_nan_stats = False,
                                                                            validate_args = True,
                                                                            name = 'distribution')
            else:
                h_dist = tfp.distributions.MultivariateNormalDiag(loc = h_mu,
                                                                  scale_diag = h_std,
                                                                  allow_nan_stats = False,
                                                                  validate_args = True,
                                                                  name = 'distribution')

            h_sample = h_dist.sample()
            h_epsilon = tf.zeros(tf.shape(h_mu))

        self.dist    = h_dist
        self.sample  = h_sample
        self.mu      = h_mu
        self.std     = h_std
        self.var     = h_var
        self.logvar  = h_logvar
        self.epsilon = h_epsilon

    def get_tensors(self):
        return [self.sample, self.mu, self.var, self.logvar, self.epsilon]



import os
import time
import logging

import itertools

# Math
import pandas as pd
import numpy as np

# Plotting
import matplotlib
import matplotlib.pyplot as plt
import matplotlib.colors as colors
from matplotlib.backends.backend_pdf import PdfPages
import seaborn as sns

# sklearn
from sklearn.decomposition import PCA
from sklearn.preprocessing import scale

# Custom
from siVAE import util
from siVAE.util import reduce_dimensions
from siVAE.model.plot import plot_images

import scanpy as sc
import gseapy as gp

def getPCALoadings(pca):
    return((pca.components_.T * np.sqrt(pca.explained_variance_)).T)


def extract_value(result_dict):
    """ """
    result  = result_dict['model']
    if 'sample' not in result_dict.keys():
        z_mu    = result['z_mu']
        z_var   = result['y_mu']
        X       = result['reconstruction'][0]
        X_recon = result['reconstruction'][1]
        values_dict = {'model': {'z_mu'   : z_mu,
                                 'z_var'  : z_var,
                                 'X'      : X,
                                 'X_recon': X_recon}}
    else:
        z_mu    = result['latent_embedding']['sample']
        z_var   = result['latent_embedding_var']['sample']
        v_mu    = result['latent_embedding']['feature']
        v_var   = result['latent_embedding_var']['feature']
        X       = result['reconstruction'][0]
        X_recon = result['reconstruction'][1]
        #
        values_dict = {'model': {'z_mu'   : z_mu,
                                 'z_var'  : z_var,
                                 'v_mu'   : v_mu,
                                 'v_var'  : v_var,
                                 'X'      : X,
                                 'X_recon': X_recon}}

        # Decoder layers in sample
        if 'decoder_layers' in result.keys():
            decoder_layers_sample = result['decoder_layers']['sample'][:-1]
            decoder_layers_feature = result['decoder_layers']['feature']
            decoder_layers_sample  = [z_mu] + decoder_layers_sample
            decoder_layers_feature = [v_mu] + decoder_layers_feature
            values_dict['model']['decoder_layers'] = {'sample' :decoder_layers_sample,
                                                      'feature':decoder_layers_feature}

    # Feature Attributions
    if 'sample_dict' in result.keys():
        sample_dict = result['sample_dict']
        if 'attributions_samples' in sample_dict.keys():
            attrb_dict = sample_dict['attributions_samples']
            values_dict['model']['Feature Attribution'] = {}
            for AE_part,FA_dict in attrb_dict.items():
                FA_scores  = FA_dict['score']
                if AE_part == 'encoder':
                    FA_scores = np.swapaxes(FA_scores,2,3)
                FA_methods = FA_dict['methods']
                values_dict['model']['Feature Attribution'][AE_part] = FA_scores
            values_dict['model']['Feature Attribution Methods'] = FA_methods

    # Sample-wise VAE
    if 'sample' in result_dict.keys():
        result_sample = result_dict['sample']
        if result_sample is not None:
            z_mu  = result_sample['z_mu']
            z_var = result_sample['z_var']
            values_dict['sample'] = {'z_mu' : z_mu,
                                     'z_var': z_var}
    # Feature-wise VAE
    if 'feature' in result_dict.keys():
        result_feature = result_dict['feature']
        if result_feature is not None:
            z_mu  = result_feature['z_mu']
            z_var = result_feature['z_var']
            values_dict['feature'] = {'z_mu' : z_mu,
                                      'z_var': z_var}

    return values_dict


def reformat_FA(scores_list,AE_part):
    """
    Reformat output of feature attribution scores into
    [FI_method, LE, sample, feature]
    """
    FA_method_list = [m for m in scores_list[0].keys()]
    scores_array = np.array([[score for score in score_dict.values()] for score_dict in scores_list])
    scores_array = np.transpose(scores_array, axes = [2,1,0,3]) # Sample x FI_method x Features x LE
    if AE_part == 'decoder':
        scores_array = np.swapaxes(scores_array,2,3)
    scores_array_FA = np.transpose(scores_array,[1,2,0,3]) # FI_method x LE x sample x Features
    return scores_array_FA,FA_method_list


def infer_FA_loadings(scores):
    """
    infer FA loadings from FA scores
    Parameters
    ----------
    scores
        array: [n_methods, n_latent_embeddings, n_samples, n_features]
    """
    loadings = []
    for scores_FA in scores:
        loadings_LE = []
        for scores_LE in scores_FA:
            pca = PCA(n_components = 1)
            FA_loadings = pca.fit_transform(scores_LE.T).T #
            loadings_LE.append(FA_loadings)
        loadings.append(loadings_LE)
    return(np.array(loadings).sum(axis=-2))


def map_with_loadings(values_dict,logdir,palette,plot_args_dict_sf,**kwargs):

    FA_methods = values_dict['model']['Feature Attribution Methods']
    z_mu = values_dict['model']['z_mu']
    X_in = values_dict['model']['X']

    for AE_part,scores in values_dict['model']['Feature Attribution'].items():

        FA_loadings = infer_FA_loadings(np.swapaxes(scores,1,2))

        for method, loadings in zip(FA_methods,FA_loadings):
            ## Map latent embedding with inferred loadings
            X = np.matmul(z_mu,loadings)

        ## Plot mapping
        data_type = AE_part
        figname = os.path.join(logdir,"Scatter-{}.pdf".format(data_type))
        labels_test = plot_args_dict_sf['sample']['labels']

        with PdfPages(figname) as pp:
            for method, loadings in zip(FA_methods,FA_loadings):
                ## Map latent embedding with inferred loadings
                X = np.matmul(z_mu,loadings)
                np.square(X - X_in).sum(1).mean()
                ## Dimensionality Reduction for plotting
                X_plot, dim_labels = reduce_dimensions(X[:5000], reduced_dimension = 2, method = 'tSNE')
                df_plot = pd.concat([pd.DataFrame(X_plot), pd.DataFrame(labels_test[:5000])], axis = 1)
                df_plot.columns = dim_labels + ["Label"]
                # Plot
                ax = sns.scatterplot(x = dim_labels[0], y = dim_labels[1], hue = "Label", data = df_plot, palette = palette, **kwargs)
                ax.legend_.remove()
                pp.savefig()
                plt.clf()

        plt.close()



def plot_FA_loadings(values_dict,logdir,ImageDims):
    """ Plot the inferred FA loadings """
    FA_methods = values_dict['model']['Feature Attribution Methods']
    for AE_part,scores in values_dict['model']['Feature Attribution'].items():
        FA_loadings = infer_FA_loadings(np.swapaxes(scores,1,2))
        images_list = FA_loadings.reshape(*FA_loadings.shape[:-1],*ImageDims)
        colnames = ['Dimensions'] + FA_methods
        rownames = ['dim-'+str(dim+1) for dim in range(len(images_list[0]))]
        _ = plot_images(images_list,colnames=colnames,rownames=rownames,cutoff=0.1)
        figname = os.path.join(logdir,"Loadings-{}.pdf".format(AE_part))
        plt.savefig(figname)
        plt.close()


def plot_siVAE_loadings(values_dict, logdir, ImageDims):
    loadings = values_dict['model']['v_mu'].T
    images_list = [loadings.reshape(*loadings.shape[:-1],*ImageDims)]
    colnames = ['Dimensions','siVAE']
    rownames = ['dim-'+str(dim+1) for dim in range(len(images_list[0]))]
    _ = plot_images(images_list,colnames=colnames,rownames=rownames,cutoff=0.1)
    figname = os.path.join(logdir,"Loadings-siVAE.pdf")
    plt.savefig(figname)
    plt.close()


def loading_analysis(values_dict, plot_args_dict_sf, logdir, dims = 2, num_genes = 20, genesets = None, **kwargs):
    """  """

    labels_gene = plot_args_dict_sf['feature']['names']
    loadings = values_dict['model']['v_mu']

    for d in range(dims):
        loading = loadings[:,d]
        ranking = np.argsort(loading)[::-1]
        ranking = np.concatenate([ranking[:num_genes],ranking[-num_genes:]])
        loading_ranked = loading[ranking]
        labels_ranked = labels_gene[ranking]

        df = pd.DataFrame({'Genes':labels_ranked,'Contribution':loading_ranked})
        g = sns.barplot(x='Genes',y='Contribution',data=df,**kwargs)
        g.set_xticklabels(labels_ranked,rotation=90,size=5)
        figname = os.path.join(logdir,"Loadings_Feature-dim_{}.pdf".format(d+1))
        plt.tight_layout()
        plt.savefig(figname)
        plt.close()

        ## Save to CSV
        ranking = np.argsort(loading)[::-1]
        loading_ranked = loading[ranking]
        labels_ranked = labels_gene[ranking]
        df = pd.DataFrame({'Genes':labels_ranked,'Contribution':loading_ranked})
        df.to_csv(os.path.join(logdir,'Loadings_Feature-dim_{}.csv'.format(d+1)),sep=',',index=False,header=True)

        ## Perform GSEA
        if genesets is not None:
            for gs_name,geneset in genesets.items():
                outdir = os.path.join(logdir,'GSEA','siVAE',gs_name,'dim-{}'.format(d))
                pre_res = gp.prerank(rnk = df, gene_sets = geneset,
                                     outdir = outdir, max_size = 200)


    X = values_dict['model']['X']
    pca = PCA(dims)
    pca.fit_transform(X)
    loadings = getPCALoadings(pca)

    labels_gene = plot_args_dict_sf['feature']['names']

    for d in range(dims):
        loading = loadings[d]
        ranking = np.argsort(loading)[::-1]
        ranking = np.concatenate([ranking[:num_genes],ranking[-num_genes:]])
        loading_ranked = loading[ranking]
        labels_ranked = labels_gene[ranking]

        df = pd.DataFrame({'Genes':labels_ranked,'Contribution':loading_ranked})
        g = sns.barplot(x='Genes',y='Contribution',data=df,**kwargs)
        g.set_xticklabels(labels_ranked,rotation=90,size=5)
        figname = os.path.join(logdir,"Loadings-PCA_{}.pdf".format(d+1))
        plt.tight_layout()
        plt.savefig(figname)
        plt.close()

        ## Save to CSV
        ranking = np.argsort(loading)[::-1]
        loading_ranked = loading[ranking]
        labels_ranked = labels_gene[ranking]
        df = pd.DataFrame({'Genes':labels_ranked,'Contribution':loading_ranked})
        df.to_csv(os.path.join(logdir,'Loadings-Feature-PC_{}.csv'.format(d+1)),sep=',',index=False,header=True)

        ## Perform GSEA
        if genesets is not None:
            for gs_name,geneset in genesets.items():
                outdir = os.path.join(logdir,'GSEA','PCA',gs_name,'dim-{}'.format(d))
                pre_res = gp.prerank(rnk = df, gene_sets = geneset,
                                     outdir = outdir, max_size = 200)


def get_feature_awareness(values_dict,plot_args_dict_sf=None,logdir=None,num_genes=20,genesets=None,**kwargs):

    x_sample = values_dict['model']['X']
    x_recon_base = x_sample.mean(0) # for centered data
    recon_loss_base = np.square(x_sample - x_recon_base)

    decoder_layers = values_dict['model']['decoder_layers']
    DL_sample  = decoder_layers['sample']
    DL_feature = decoder_layers['feature']
    x_recon_list = [np.matmul(score,FL.transpose()) for FL,score in zip(DL_feature,DL_sample)]
    recon_loss_model = np.square(x_sample - x_recon_list)

    FeatureAwareness = -(recon_loss_model - recon_loss_base)

    if plot_args_dict_sf is not None and logdir is not None:

        labels_gene = plot_args_dict_sf['feature']['names']

        for d,FeatAw in enumerate(FeatureAwareness):
            ## Plot
            loading = FeatAw.mean(0)
            ranking = np.argsort(loading)[::-1]
            ranking = np.concatenate([ranking[:num_genes],ranking[-num_genes:]])
            loading_ranked = loading[ranking]
            labels_ranked = labels_gene[ranking]

            df = pd.DataFrame({'Genes':labels_ranked,'Contribution':loading_ranked})
            g = sns.barplot(x='Genes',y='Contribution',data=df,**kwargs)
            _ = g.set_xticklabels(labels_ranked,rotation=90,size=5)
            figname = os.path.join(logdir,"FeatureAwareness-TopFeature-layer_{}.pdf".format(d+1))
            plt.tight_layout()
            plt.savefig(figname)
            plt.close()

            ## Save to CSV
            loading = FeatAw.mean(0)
            ranking = np.argsort(loading)[::-1]
            loading_ranked = loading[ranking]
            labels_ranked = labels_gene[ranking]
            df = pd.DataFrame({'Genes':labels_ranked,'Contribution':loading_ranked})
            df.to_csv(os.path.join(logdir,'FA_TopFeature-layer_{}.csv'.format(d+1)),sep=',',index=False,header=True)

            ## Perform GSEA
            if genesets is not None:
                for gs_name,geneset in genesets.items():
                    outdir = os.path.join(logdir,'GSEA','FeatureAwareness',gs_name,'dim-{}'.format(d))
                    pre_res = gp.prerank(rnk = df, gene_sets = geneset,
                                         outdir = outdir, max_size = 200)

    return FeatureAwareness


def save_losses(result_dict,logdir):
    losses = result_dict['model']['losses']
    df = pd.DataFrame(losses)
    df.to_csv(os.path.join(logdir,'losses.csv'),sep=',',index=False,header=True)

    if 'sample' in result_dict.keys():
        if result_dict['sample'] is not None:
            losses = result_dict['sample']['losses']
            df = pd.DataFrame(losses)
            df.to_csv(os.path.join(logdir,'losses_sample.csv'),sep=',',index=False,header=True)

    if 'feature' in result_dict.keys():
        if result_dict['feature'] is not None:
            losses = result_dict['feature']['losses']
            df = pd.DataFrame(losses)
            df.to_csv(os.path.join(logdir,'losses_feature.csv'),sep=',',index=False,header=True)


def get_feature_awareness_pca(values_dict,plot_args_dict_sf=None,logdir=None,num_genes=20,**kwargs):

    x_sample = values_dict['model']['X']
    x_recon_base = x_sample.mean(0) # for centered data
    recon_loss_base = np.square(x_sample - x_recon_base)
    LE_dim  = values_dict['model']['z_mu'].shape[-1]

    pca = PCA(LE_dim)
    x_pca   = pca.fit_transform(x_sample)
    x_recon = pca.inverse_transform(x_pca)
    recon_loss_model = np.square(x_sample - x_recon)
    FeatureAwareness = -(recon_loss_model - recon_loss_base)

    if plot_args_dict_sf is not None and logdir is not None:

        labels_gene = plot_args_dict_sf['feature']['names']

        ## Plot
        loading = FeatureAwareness.mean(0)
        ranking = np.argsort(loading)[::-1]
        ranking = np.concatenate([ranking[:num_genes],ranking[-num_genes:]])
        loading_ranked = loading[ranking]
        labels_ranked = labels_gene[ranking]
        df = pd.DataFrame({'Genes':labels_ranked,'Contribution':loading_ranked})
        g = sns.barplot(x='Genes',y='Contribution',data=df,**kwargs)
        _ = g.set_xticklabels(labels_ranked,rotation=90,size=5)
        figname = os.path.join(logdir,"FA_TopFeature-PCA_{}.pdf".format(LE_dim))
        plt.tight_layout()
        plt.savefig(figname)
        plt.close()

        ## Save to CSV
        loading = FeatureAwareness.mean(0)
        ranking = np.argsort(loading)[::-1]
        loading_ranked = loading[ranking]
        labels_ranked = labels_gene[ranking]
        df = pd.DataFrame({'Genes':labels_ranked,'Contribution':loading_ranked})
        df.to_csv(os.path.join(logdir,'FA_TopFeature-PCA_{}.csv'.format(LE_dim)),sep=',',index=False,header=True)

    return FeatureAwareness


def plot_latent_embeddings(values_dict, plot_args_dict_sf, palette, method_dim_reds = ['PCA','tSNE'], logdir='', multidimension=False, show_legend=True,**kwargs):

    z_mu = values_dict['model']['z_mu']
    v_mu = values_dict['model']['v_mu']

    X_dict = {'Sample-Wise' : z_mu,
              'Feature-wise': v_mu,
              'Reconstruction' : values_dict['model']['X_recon'],
              'Original'       : values_dict['model']['X']}

    if 'sample' in values_dict.keys():
        X_dict['Sample-Wise-pre'] = values_dict['sample']['z_mu']

    if 'feature' in values_dict.keys():
        X_dict['Feature-Wise-pre'] = values_dict['feature']['z_mu']

    for data_type, X in X_dict.items():

        if data_type in ['Sample-Wise','Sample-Wise-pre','Original','Reconstruction']:
            labels_test = plot_args_dict_sf['sample']['labels']

        else:
            labels_test = plot_args_dict_sf['feature']['labels']

        for method_dim_red in method_dim_reds:
            # dataframe with reduced dimensions
            X_plot, dim_labels = reduce_dimensions(X, reduced_dimension = 2, method = method_dim_red)
            df_plot = pd.concat([pd.DataFrame(X_plot), pd.DataFrame(labels_test)], axis = 1)
            df_plot.columns = dim_labels + ["Label"]
            # Plot
            ax = sns.scatterplot(x = dim_labels[0], y = dim_labels[1], hue = "Label", data = df_plot, palette = palette, **kwargs)
            if not show_legend:
                ax.legend_.remove()
            figname = os.path.join(logdir,"Scatter-{}_{}.pdf".format(data_type,method_dim_red))
            plt.savefig(figname)
            plt.close()

        if multidimension and data_type == 'Sample-Wise' and X.shape[-1] < 5:

            method_dim_red = 'multidimension'
            figname = os.path.join(logdir,"Scatter-{}_{}.pdf".format(data_type,method_dim_red))

            with PdfPages(figname) as pp:
                for i,j in itertools.combinations(range(X.shape[-1]),2):
                    X_plot = X[:,[i,j]]
                    dim_labels = ['dim-{}'.format(i),'dim-{}'.format(j)]
                    df_plot = pd.concat([pd.DataFrame(X_plot), pd.DataFrame(labels_test)], axis = 1)
                    df_plot.columns = dim_labels + ["Label"]
                    # Plot
                    ax = sns.scatterplot(x = dim_labels[0], y = dim_labels[1], hue = "Label", data = df_plot, palette = palette, **kwargs)
                    if not show_legend:
                        ax.legend_.remove()
                    pp.savefig()
                    plt.clf()
            plt.close()

        ax = sns.scatterplot(x = dim_labels[0], y = dim_labels[1], hue = "Label", data = df_plot, palette = palette, **kwargs)
        lgd = plt.legend(bbox_to_anchor=(1.05, 1), loc=2, borderaxespad=0.)
        plt.tight_layout()
        figname = os.path.join(logdir,"Scatter-legend-{}.pdf".format(data_type))
        plt.savefig(figname, bbox_extra_artists=(lgd,), bbox_inches='tight')
        plt.close()


def calculate_PVE(X_in, X_predicted):
    var_orig = np.square(X_in.std(0))
    X_resids = X_in - X_predicted
    var_resid = np.square(X_resids.std(-2))
    PVE = (var_orig.sum(-1) - var_resid.sum(-1)) / var_orig.sum(-1)
    return(PVE)


def compare_loadings(values_dict,logdir,PCA_loadings=None):

    siVAE_loadings = values_dict['model']['v_mu'].T

    dims = siVAE_loadings.shape[0]

    if dims < 5:

        if PCA_loadings is None:
            X = values_dict['model']['X']
            pca = PCA(siVAE_loadings.shape[0])
            pca.fit(X)
            PCA_loadings = getPCALoadings(pca)

        figname = os.path.join(logdir,'loadings_comparison_scatter.pdf')
        with PdfPages(figname) as pp:
            for i,j in itertools.product(range(dims),repeat=2):
                df = pd.DataFrame({'siVAE':siVAE_loadings[i],'PCA':PCA_loadings[j]})
                ax = sns.scatterplot(data=df,x='siVAE',y='PCA')
                plt.title('siVAE_dim-{} vs PC-{}'.format(i+1,j+1))
                pp.savefig()
                plt.clf()
        plt.close()


def compare_FeatureAwareness(FA_siVAEs,FA_pca,logdir):
    figname = os.path.join(logdir,'FA_comparison_scatter.pdf')
    with PdfPages(figname) as pp:
        for ii,FA_siVAE in enumerate(FA_siVAEs):
            df = pd.DataFrame({'siVAE':FA_siVAE.mean(0),'PCA':FA_pca.mean(0)})
            ax = sns.scatterplot(data=df,x='siVAE',y='PCA')
            corr = np.corrcoef(FA_siVAE,FA_pca)[0,1]
            plt.title("siVAE l-{} vs PCA: {:.3f}".format(ii,corr))
            pp.savefig()
            plt.clf()
    plt.close()


def PVE_analysis(values_dict,save_result=False,logdir=""):

    X_orig = values_dict['model']['X']
    var_orig = np.square(X_orig.std(0))
    decoder_layers = values_dict['model']['decoder_layers']
    DL_sample  = decoder_layers['sample']
    DL_feature = decoder_layers['feature']
    PVE_layers = []

    for ii,(FL,score) in enumerate(zip(DL_feature,DL_sample)):
        X_recon = np.matmul(score,FL.transpose())
        PVE_layers.append(calculate_PVE(X_orig,X_recon))

        if ii == 0:
            dims = score.shape[1]
            X_dims = np.array([np.matmul(score[:,[d]],FL[:,[d]].transpose()) for d in range(dims)])
            PVE_dims = calculate_PVE(X_orig,X_dims)

    if save_result:
        np.savetxt(os.path.join(logdir,'PVE_dims.csv'),PVE_dims,delimiter=',')
        np.savetxt(os.path.join(logdir,'PVE_layers.csv'),PVE_layers,delimiter=',')

    return(np.array(PVE_layers),PVE_dims)


def recode_embeddings(values_dict, plot_args_dict_sf, ImageDims, n_pc = 3,
                      logdir = ''):
    ## Create subsets
    labels = plot_args_dict_sf['sample']['labels']
    names = np.unique(labels).tolist()
    masks = [np.isin(labels,ii) for ii in np.unique(labels)]

    score_list = values_dict['model']['decoder_layers']['sample']
    loadings_list = values_dict['model']['decoder_layers']['feature']
    decomposition_dict = {"l_{}".format(l+1): (score,loadings.T) for l,(score,loadings) in enumerate(zip(score_list,loadings_list))}
    X = values_dict['model']['X']

    ## 1. Visualize Recoded Embeddings
    for level, (score,loadings) in decomposition_dict.items():
        if score.shape[-1] >= 5:
            images_list = []
            for mask in masks:
                score_masked = score[mask]

                ## Recode
                pca = PCA(n_components = n_pc).fit(score_masked)
                recoded_score = pca.transform(score_masked)
                recoded_loadings = (pca.components_.T * np.sqrt(pca.explained_variance_)).T
                recoded_pcs = np.matmul(recoded_loadings,loadings)
                images_list.append(recoded_pcs)

            ## Plot
            images_list = np.swapaxes(np.array(images_list),0,1)
            images_list = images_list.reshape([*images_list.shape[:-1],*ImageDims])
            colnames = ['Label'] + ['pc-'+str(ii+1) for ii in range(n_pc)]
            rownames = names
            _ = plot_images(images_list, colnames = colnames, rownames=rownames, cutoff=0.01)
            figname = os.path.join(logdir,'Recoded_embeddings-{}.pdf'.format(level))
            plt.savefig(figname)
            plt.close()

    ## 2. Visualize the changes along the pc-1
    for level, (score,loadings) in decomposition_dict.items():
        if level == "l_1":
            if score.shape[-1] >= 5:
                images_list = []
                for name,mask in zip(names,masks):
                    score_masked = score[mask]
                    X_masked     = X[mask]

                    ## Recode
                    pca = PCA(n_components = n_pc).fit(score_masked)
                    recoded_score = pca.transform(score_masked)
                    recoded_loadings = (pca.components_.T * np.sqrt(pca.explained_variance_)).T
                    recoded_pcs = np.matmul(recoded_loadings,loadings)

                    lim = 1
                    idxrange = np.where(np.all([recoded_score[:,1] > -lim, recoded_score[:,1] < lim], axis=0))[0]
                    idxranked = np.argsort(recoded_score[idxrange,0])
                    idxchosen = np.round(np.linspace(0, len(idxranked) - 1, 10)).astype(int)
                    idx_use_siVAE = idxrange[idxranked[idxchosen]]

                    ## PCA recoding
                    pca = PCA(n_components = n_pc,whiten=True)
                    PCA_score = pca.fit_transform(scale(X_masked))
                    idxrange = np.where(np.all([PCA_score[:,1] > -lim, PCA_score[:,1] < lim], axis=0))[0]
                    idxranked = np.argsort(PCA_score[idxrange,0])
                    idxchosen = np.round(np.linspace(0, len(idxranked) - 1, 10)).astype(int)
                    idx_use_PCA = idxrange[idxranked[idxchosen]]

                    ## Plot
                    images_list = [idx_use_siVAE,idx_use_PCA]
                    images_list = [[X_masked[ii] for ii in idx] for idx in images_list]
                    images_list = [[im.reshape(*ImageDims) for im in images] for images in images_list]
                    images_list = [[images_list[icol][irow] for icol in range(len(images_list))] for irow in range(len(images_list[0]))]
                    rownames = ['siVAE','original']
                    colnames = ['']+['Image-{}'.format(ii+1) for ii in range(len(images_list))]
                    _ = plot_images(np.array(images_list), colnames = colnames, rownames=rownames, cutoff=0.01)
                    figname = os.path.join(logdir,'Recoded_embeddings_label-{}.pdf'.format(name))
                    plt.savefig(figname)
                    plt.close()


def plot_feature_awareness(result_dict,scaler,ImageDims,plot_args_dict_sf,logdir):
    labels = plot_args_dict_sf['sample']['labels']
    result = result_dict['model']
    x_sample = result['sample_dict']['input']
    x_recon  = result['sample_dict']['output']
    x_sample_inversed = scaler.inverse_transform(x_sample)
    x_recon_inversed  = scaler.inverse_transform(x_sample)
    x_sample_image    = x_sample_inversed.reshape(*x_sample_inversed.shape[:-1],*ImageDims)
    x_recon_image     = x_recon_inversed.reshape(*x_recon_inversed.shape[:-1],*ImageDims)

    x_recon_base = 0 # for centered data
    recon_loss_base = np.square(x_sample - x_recon_base) + 1e-3

    recon_loss_list = []
    x_recon_list = result_dict['model']['sample_dict']['output_hl'] + [x_recon]

    recon_loss_list = [np.square(x_sample - x_r) for x_r in x_recon_list]

    recon_loss_image_list = []
    FeatureAwareness_list = []

    for FS in recon_loss_list:
        ##
        FS_image = FS.reshape(len(FS),*ImageDims)
        FS_image = FS_image.sum(-1) # Sum over the color channels
        FS_image = FS_image.reshape(len(FS_image),*ImageDims[:-1],1)
        recon_loss_image_list.append(FS_image)
        ##
        FS_centered = FS - recon_loss_base
        FS_centered_image = FS_centered.reshape(len(FS_centered),*ImageDims)
        FS_centered_image = FS_centered_image.sum(-1) # Sum over the color channels
        FS_centered_image = FS_centered_image.reshape(len(FS_centered_image),*ImageDims[:-1],1)
        FeatureAwareness_list.append(FS_centered_image)

    images_list = [x_sample_image] + FeatureAwareness_list + [x_recon_image]
    recon_names = ['l-'+str(ii+1) for ii,_ in enumerate(FeatureAwareness_list)]
    colnames = ['Label','Original'] + recon_names + ['Reconstruction']
    rownames = np.unique(labels)
    idx_sample = np.array([int(len(x_sample)/len(np.unique(labels)) *ii) for ii,_ in enumerate(np.unique(labels))])
    images_list = [images[idx_sample] for images in images_list]
    _ = plot_images(images_list, colnames = colnames, rownames = rownames, cutoff = 0.1)
    figname = os.path.join(logdir,'FeatureAwareness.pdf')
    plt.savefig(figname)
    plt.close()

## =============================================================================
## Custom
## =============================================================================

def infer_loadings(values_dict, grids = None, method = 'linear'):
    """
    Infer loadings based on additional information
    If method is linear, use linear algebra to solve for loadings
    If method is grid, use sampling to solve for loadings
    """

    if type(method) is list:

        inferred_loadings = np.array([infer_loadings(values_dict,grid,m) for m in method])

    else:
        z_mu = values_dict['model']['z_mu']

        if method == 'linear':

            X_recon = values_dict['model']['X_recon']

            part1 = np.linalg.inv(np.matmul(z_mu.transpose(),z_mu))
            part2 = np.matmul(z_mu.transpose(),X_recon)

            inferred_loadings = np.matmul(part1,part2)

        elif method == 'grid':

            if grids is None:
                raise Exception('Input grid must be specified')

            n_LE = z_mu.shape[1]

            grids = np.split(grids,n_LE)

            inferred_loadings = []

            for grid in grids:
                diff = grid[1:] - grid[:-1]
                inferred_loading1 = diff.mean(0)
                pca = PCA(n_components = 1)
                inferred_loading2 = pca.fit_transform(grid.transpose())[:,0]
                inferred_loadings.append(np.array([inferred_loading1]))

            inferred_loadings = np.array(inferred_loadings)[:,0]

    return inferred_loadings



import os
import time
import logging

# Tensorflow
import tensorflow as tf
import tensorflow_probability as tfp
tf.compat.v1.disable_v2_behavior()

import numpy as np
import pandas as pd

from .layer import layer
from .layer import KL_divergence_normal
from .layer import create_variance
from .layer import Distribution

from .util import batch_handler

class AutoEncoder(object):

    def __init__(self, logdir_tf = ".", data_handler = None, X_dim = None,
                 h_dims = None, LE_dim = None, architecture = None,
                 iter = 5000, mb_size = 1000, learning_rate = 1e-3, l1_scale = 1e-50, l2_scale = 0,
                 early_stopping = 0, tolerance = 0, min_early_stopping = 0,
                 activation_fun = tf.nn.relu, random_seed = 0, log_frequency = 100,
                 batch_norm = False, keep_prob = None,
                 dataAPI = False, tensorboard = False, metadata = False,
                 config = None, validation_split = 0,
                 decay_rate = 1, decay_steps = 1000, save_recon = True,
                 save_LE = True, isVAE = True, beta = 1,
                 decoder_var = 'scalar', save_W = True,
                 name = "", var_dependency = True, optimizer_type = 'adam',
                 kernel_initializer = tf.contrib.layers.xavier_initializer(uniform = False),
                 X_mu_use_bias = True, beta_warmup = 0.5, y_var_type = 'deterministic',
                 classifier = False, attributions_dict = {}):

        self.iterator = None
        self.isVAE = isVAE
        self.decoder_var = decoder_var
        self.y_var_type = y_var_type
        self.var_dependency = var_dependency
        self.X_mu_use_bias = X_mu_use_bias
        self.beta_warmup = beta_warmup
        self.beta = beta
        if not self.isVAE and decoder_var != 'deterministic':
            raise Exception("When isVAE is False, decoder_var must be set to deterministic")

        # NN structure
        self.data_handler           = data_handler
        self.h_dims                 = self.data_handler.h_dims
        self.X_dim                  = self.data_handler.X.shape[-1]
        self.X_target_dim           = self.data_handler.y.shape[-1]
        self.index_latent_embedding = self.data_handler.index_latent_embedding
        shape_X    = list(self.data_handler.X.shape)
        shape_y    = list(self.data_handler.y.shape)
        shape_X[0] = None
        shape_y[0] = None
        self.X_shape        = shape_X
        self.X_target_shape = shape_y

        ## Recalculate h_dims if h_dims = 0
        self.save_LE = save_LE
        self.save_W = save_W
        self.save_recon = save_recon

        ## Convert h_dims
        self.process_h_dims()

        logging.info("h_dims: {}".format(self.h_dims))

        self.name = name

        ## Set activation function
        self.activation_fun = activation_fun

        ## Training Parameters
        self.mb_size = mb_size
        self.iter = iter
        self.keep_prob = keep_prob

        self.kernel_initializer = kernel_initializer
        self.optimizer_type = optimizer_type

        ## Hyperparameters
        self.learning_rate = learning_rate
        self.l2_scale = l2_scale
        self.l1_scale = l1_scale
        self.random_seed = random_seed

        self.batch_norm = batch_norm

        self.placeholder_reg = {}

        self.feed_dict = {}
        self.logdir_tf = logdir_tf

        self.tensorboard = tensorboard
        self.metadata = metadata
        self.validation_split = validation_split
        self.tolerance = tolerance
        self.summary_list = []

        self.decay_steps = decay_steps
        self.decay_rate = decay_rate


        if self.metadata and not self.tensorboard:
            logging.info("Tensorboard has been switched to true as metadata requires tensorboard")
            self.tensorboard = True

        self.log_frequency = log_frequency
        self.dataAPI = dataAPI
        self.config = config

        self.early_stopping = early_stopping
        self.min_early_stopping = np.max([min_early_stopping, self.beta_warmup])
        self.iter_min_early_stopping = int(self.iter * self.min_early_stopping)

        for attribution,value in attributions_dict.items():
            setattr(self,attribution,value)

    def process_h_dims(self):
        """ Process number of nodes for hidden layers """

        if len(self.h_dims) != 0:

            if np.any(np.array(self.h_dims[:-1]) == 0):

                h_dims_old = [self.X_dim] + self.h_dims
                h_dims_new = h_dims_old
                logging.info("Converting h_dims: ".format(h_dims_old))

                for ii, h_dim in enumerate(h_dims_old[:-1]):

                    if h_dim == 0:
                        logging.info(ii)
                        end = ii + 1

                        if h_dims_old[end] != 0:
                            new_values = np.linspace(h_dims_old[start], h_dims_old[end], end - start + 1).astype(int)
                            h_dims_new[start+1:end] = new_values[1:-1]

                    else:
                        start = ii

                logging.info("h_dims_new: ".format(h_dims_new))
                self.h_dims = h_dims_new[1:]

            self.h_dims[-1] = self.X_target_dim


    def build_model(self, reset_graph = True, graph = None):
        """ """

        if reset_graph:
            tf.reset_default_graph()

        if graph is None:
            self.Graph = tf.get_default_graph()
        else:
            self.Graph = graph

        tf.set_random_seed(self.random_seed)

        with tf.variable_scope(self.name):
            with tf.name_scope('regularization'):
                self.create_regularization()

            with tf.variable_scope('feed'):
                self.create_feed(dataAPI = self.dataAPI)

            with tf.variable_scope('model'):
                self.build_basic_network()


    def create_feed(self, dataAPI = False, batch = None, epoch = None, shuffle = None):
        """ y = X_reconstructed for variational autoencoder """

        if self.data_handler.iterator is not None:
            X = self.data_handler.iterator.get_next()[0]
            self.X = X
            self.X_target = X
        else:
            if not hasattr(self, 'X'):
                with tf.name_scope('X_placeholder'):
                    self.X = tf.placeholder(tf.float32, shape=self.X_shape, name = 'X')

            if not hasattr(self, 'X_target'):
                with tf.name_scope('X_target_placeholder'):
                    self.X_target = tf.placeholder(tf.float32, shape=self.X_target_shape, name = 'X_target')

        if dataAPI:
            self.dataAPI_placeholders = [self.X,self.X_target]

            ## dummy_dataset created as placeholder
            dummy_dataset = tf.compat.v1.data.Dataset.from_tensor_slices((self.X,self.X_target)).batch(10000)

            iterator = tf.compat.v1.data.Iterator.from_structure(dummy_dataset.output_types,
                                                                 dummy_dataset.output_shapes)

            self.X, self.X_target = iterator.get_next()

            self.iterator = iterator

        if not hasattr(self, 'initial_feed'):
            self.initial_feed = self.X


    def create_regularization(self):
        """ Create regularizations for the model """

        if self.batch_norm:
            self.is_train = tf.placeholder_with_default(False, shape=(), name = "is_train")
            self.placeholder_reg["batch_norm"] = self.is_train

        if self.keep_prob is not None:
            self.prob = tf.placeholder_with_default(np.array(1).astype("float32"),
                                                    shape=(), name = 'prob')
            self.placeholder_reg["drop"] = self.prob

        self.do_sample = tf.placeholder_with_default(True, shape=(), name = "do_sample")


    def create_decoder_loss(self):
        """ Create decoder loss """

        with tf.name_scope('mean_squared_error'):
            self.recon_loss_ae = self.mean_squared_error(self.X_mu, self.X_target)

        with tf.name_scope('log_likelihood'):
            if self.isVAE and self.decoder_var != 'deterministic':
                self.decoder_loss_old = self.X_logvar / 2 + self.recon_loss_ae / self.X_var / 2 + tf.constant(0.5 * np.log(2. * np.pi), dtype = self.X_var.dtype)
                self.decoder_loss_mse    = tf.square(self.X_mu - self.X_target)
                self.decoder_loss_mse_scaled = self.recon_loss_ae / self.X_var / 2
                self.decoder_loss_logvar = self.decoder_loss_mse * 0 + self.X_logvar / 2
                self.decoder_loss = -self.X_dist.dist.log_prob(self.X_target)
            else:
                self.decoder_loss = tf.reduce_sum(self.recon_loss_ae,-1)


    def build_encoder(self, h_temp, h_dims, kernel_initializer, variable_scope = 'Encoder', isVAE = False):
        """ Build encoder """

        logging.info("Building Encoder")

        l2_scale_encoder = self.l2_scale
        l1_scale_encoder = self.l1_scale

        with tf.variable_scope(variable_scope):
            for ii, h_dim in enumerate(h_dims):
                logging.info("building layer {} with {} nodes".format(ii,h_dim))
                with tf.variable_scope("hidden_layer_{}".format(ii)):
                    if ii < len(h_dims) - 1:
                        ## Create hidden layers
                        h_temp,_,_ = layer(h = h_temp,
                                           dim = h_dim,
                                           fun = self.activation_fun,
                                           l2_scale = l2_scale_encoder,
                                           l1_scale = l1_scale_encoder,
                                           kernel_initializer = kernel_initializer,
                                           **self.placeholder_reg)
                    else:

                        ## create latent embedding layer
                        if isVAE:
                            var_type = 'diagonal'
                        else:
                            var_type = 'deterministic'

                        prob_layer = Distribution(h = h_temp,
                                                  dim = h_dim,
                                                  fun = None,
                                                  l2_scale = l2_scale_encoder,
                                                  l1_scale = l1_scale_encoder,
                                                  var_type = 'diagonal')

                self.hidden_layers.append(h_temp)

        return prob_layer


    def build_decoder(self, h_dims, h_temp, l1_scale, l2_scale, kernel_initializer, variable_scope = 'decoder'):

        logging.info("Building " + variable_scope + "with dims {}".format(h_dims))

        self.decoder_layers = []

        with tf.variable_scope(variable_scope, reuse = tf.AUTO_REUSE):

            # self.decoder_variable_scope = tf.get_variable_scope().name

            for ii, h_dim in enumerate(h_dims):

                with tf.variable_scope("hidden_layer_{}".format(ii)):

                    if ii < len(h_dims) - 2:

                        ## Hidden layers
                        fun = self.activation_fun

                        h_temp,_,_ = layer(h = h_temp, dim = h_dim, fun = fun,
                                           l2_scale = l2_scale, l1_scale = l1_scale,
                                           kernel_initializer = kernel_initializer)

                    elif ii == len(h_dims) - 2:
                        ## Last layer before final layer

                        prob_layer  = Distribution(h   = h_temp,
                                                   dim = h_dim,
                                                   fun = None,
                                                   var_type = self.y_var_type,
                                                   l2_scale = l2_scale,
                                                   l1_scale = l1_scale)

                        self.y_dist = prob_layer
                        self.y_sample, self.y_mu, self.y_var, self.y_logvar, self.y_eps = prob_layer.get_tensors()
                        h_temp = self.y_sample

                    elif ii == len(h_dims) - 1:

                        ## If there is no y layer, set y to z
                        if len(h_dims) == 1:
                            self.y_dist = self.z_dist
                            self.y_sample = self.z_sample
                            self.y_mu = self.z_mu
                            self.y_var = self.z_var
                            self.y_logvar = self.z_logvar
                            self.y_eps    = self.z_eps

                        with tf.variable_scope('output', reuse = tf.AUTO_REUSE):

                            self.output_variable_scope = tf.get_variable_scope().name

                            prob_layer,W,b = self.build_output_layer(h_temp, h_dim = h_dim,
                                                                     fun = self.activation_fun,
                                                                     var_type = self.decoder_var,
                                                                     var_dependency = self.var_dependency,
                                                                     l2_scale = l2_scale,
                                                                     l1_scale = l1_scale,
                                                                     use_bias = self.X_mu_use_bias)

                        self.X_dist = prob_layer
                        self.X_sample, self.X_mu, self.X_var, self.X_logvar, self.X_eps = prob_layer.get_tensors()
                        self.X_sample = tf.cond(self.do_sample, lambda: self.X_sample, lambda: self.X_mu, 'X_sample_cond')
                        self.W = W
                        self.b = b
                        h_temp = self.X_mu

                self.decoder_layers.append(h_temp)
                self.hidden_layers.append(h_temp)
                self.summary_list.append(tf.summary.histogram("hidden_layer_{}", h_temp))

        return prob_layer


    def build_output_layer(self, h_temp, fun, var_type, var_dependency, l2_scale, l1_scale, use_bias,
                           h_dim = None, W = None, b = None, h_mu = None, h_std = None, custom = False, **kwargs):
        """"""

        h_mu, W, b = layer(h_temp, h_dim, fun = None,
                           name = "{}_{}".format('X','mu'),
                           l1_scale = l1_scale, l2_scale = l2_scale,
                           custom = custom, use_bias = use_bias,
                           W = W, b = b, **kwargs)


        with tf.variable_scope('Distribution'):
            prob_layer  = Distribution(h = h_temp,
                                       h_mu = h_mu,
                                       h_std = h_std,
                                       dim = h_dim,
                                       fun = None,
                                       var_type = var_type,
                                       var_dependency = var_dependency,
                                       l2_scale = l2_scale,
                                       l1_scale = l1_scale,
                                       **kwargs)

        return prob_layer, W, b


    def mean_squared_error(self, target, reconstruction):
        """ calculate mean squared error between target and reconstruction """
        return tf.square(target - reconstruction)


    def build_basic_network(self):
        """ Build network """
        logging.info("Building Basic Network")

        with tf.variable_scope('NN'):
            self.X_out = self.predict_X(self.initial_feed)

        with tf.name_scope('losses'):

            ## Autoencoder
            logging.info("recon_loss_ae")
            self.create_decoder_loss()
            logging.info(self.recon_loss_ae)
            logging.info("decoder_loss")
            logging.info(self.decoder_loss)

            # VAE
            self.create_loss()


    def create_loss(self):
        """ """
        logging.info("Creating loss")
        logging.info("GraphKeys Losses")
        logging.info(tf.get_collection(tf.GraphKeys.LOSSES))

        ## Set up optimizer
        with tf.name_scope("global_steps"):

            self.global_step = tf.Variable(0, name='global_step', trainable=False)

            self.decayed_learning_rate= tf.train.exponential_decay(learning_rate = self.learning_rate,
                                                                   global_step   = self.global_step,
                                                                   decay_steps   = self.decay_steps,
                                                                   decay_rate    = self.decay_rate,
                                                                   staircase     = False)
            if self.isVAE:
                with tf.name_scope("beta_scaled"):

                    self.iter_beta_full = self.iter * self.beta_warmup

                    if self.beta_warmup != 0:
                        self.beta_scale = tf.clip_by_value(tf.cast(self.global_step, dtype = tf.float32) / (self.iter_beta_full), 0, 1)
                    else:
                        self.beta_scale = 1

                    self.beta_scaled = self.beta * self.beta_scale
                    self.summary_list.append(tf.summary.scalar('beta_scaled', self.beta_scaled))


        if self.isVAE:
            logging.info("KL_loss_ae")
            logging.info("beta: {}".format(self.beta))
            # self.KL_loss = self.create_KL_loss(beta = self.beta)
            self.KL_loss = self.beta_scaled * tfp.distributions.kl_divergence(self.z_dist.dist,
                                                                           self.z_prior,
                                                                           allow_nan_stats = False,
                                                                           name = 'z_KL_divergence')

            logging.info(self.KL_loss)

        ## l2_loss
        self.reg_loss = tf.reduce_sum(tf.losses.get_regularization_losses(scope = self.name))
        logging.info(" ")
        logging.info("reg_loss: {}".format(self.reg_loss.shape))
        logging.info(self.reg_loss)
        logging.info(tf.losses.get_regularization_losses(scope = self.name))

        ## recon_loss
        self.recon_loss_scalar = tf.reduce_mean(tf.reduce_sum(self.recon_loss_ae,-1))
        logging.info("recon_loss_scalar: {}".format(self.recon_loss_scalar.shape))
        logging.info(self.recon_loss_scalar)
        logging.info(self.recon_loss_ae)

        ## decoder_loss
        self.decoder_loss_scalar = tf.reduce_mean(self.decoder_loss)
        tf.losses.add_loss(self.decoder_loss_scalar)
        logging.info("decoder_loss_scalar: {}".format(self.decoder_loss_scalar.shape))
        logging.info(self.decoder_loss_scalar)
        logging.info(self.decoder_loss)

        ## KL_loss
        if self.isVAE:
            self.KL_loss_scalar = tf.reduce_mean(self.KL_loss)
            tf.losses.add_loss(self.KL_loss_scalar)
            logging.info("KL_loss: {}".format(self.KL_loss_scalar.shape))
            logging.info(self.KL_loss_scalar)
            logging.info(self.KL_loss)

        ## total_loss
        self.total_loss = tf.losses.get_total_loss(name = 'total_loss',
                                                   add_regularization_losses = True,
                                                   scope = self.name)
        logging.info("total_loss: {}".format(self.total_loss.shape))
        logging.info(self.total_loss)

        ## Log all to tensorboard
        self.losses = [self.total_loss, self.reg_loss, self.recon_loss_scalar, self.decoder_loss_scalar]
        self.losses_name = ['Total loss', "reg loss", "recon loss", "decoder loss"]

        if self.isVAE:
            self.losses += [self.KL_loss_scalar]
            self.losses_name += ['KL_loss']

        if self.tensorboard:
            for loss, loss_name in zip(self.losses, self.losses_name):
                logging.info("Adding {} to tensorboard".format(loss_name))
                self.summary_list.append(tf.summary.scalar(loss_name, loss))

        self.theta = tf.get_collection(tf.GraphKeys.TRAINABLE_VARIABLES, scope = self.name)
        self.theta_encoder = tf.get_collection(tf.GraphKeys.TRAINABLE_VARIABLES, scope = os.path.join(self.name,"NN/masked_X/Encoder"))
        self.theta_decoder = tf.get_collection(tf.GraphKeys.TRAINABLE_VARIABLES, scope = os.path.join(self.name,"NN/masked_X/Decoder"))

        logging.info("thetas")
        logging.info(self.theta)
        logging.info("theta_encoder")
        logging.info(self.theta_encoder)
        logging.info("theta_decoder")
        logging.info(self.theta_decoder)

        update_ops = tf.get_collection(tf.GraphKeys.UPDATE_OPS, scope = self.name)

        logging.info("Build train_op")

        with tf.name_scope("train_op"):

            if self.optimizer_type == 'adam':
                self.optimizer = tf.train.AdamOptimizer(self.decayed_learning_rate)

            with tf.control_dependencies(update_ops):
                optimizer = self.optimizer.minimize(loss = self.total_loss,
                                                    global_step = self.global_step,
                                                    var_list = self.theta)
                self.train_op = optimizer

        logging.info("train_op:")
        logging.info(self.train_op)

        if self.tensorboard:
            self.merge = tf.summary.merge(self.summary_list)
            self.losses.insert(0,self.merge)


    def predict_X(self, h_temp):
        """ Build network that takes in masked X then returns Y """

        self.hidden_layers = []

        h_dims_encoder = self.h_dims[:self.index_latent_embedding + 1]
        self.h_dims_encoder = h_dims_encoder
        h_dims_decoder = self.h_dims[self.index_latent_embedding + 1:]
        self.h_dims_decoder = h_dims_decoder

        ## Encoder
        prob_layer = self.build_encoder(h_temp = h_temp, h_dims = h_dims_encoder,
                                        kernel_initializer = self.kernel_initializer,
                                        variable_scope = 'Encoder', isVAE = self.isVAE)

        self.z_dist = prob_layer
        self.z_sample, self.z_mu, self.z_var, self.z_logvar, self.z_eps = prob_layer.get_tensors()
        self.z_sample = tf.cond(self.do_sample, lambda: self.z_sample, lambda: self.z_mu, 'z_sample_cond')
        self.z_prior = tfp.distributions.MultivariateNormalDiag(loc = tf.zeros(int(self.z_sample.shape[-1])),
                                                                scale_diag = tf.ones(int(self.z_sample.shape[-1])))

        self.latent_embedding = self.z_mu

        if self.tensorboard:
            with tf.name_scope("z"):
                self.summary_list.append(tf.summary.histogram("z_sample", self.z_sample[0]))
                if self.isVAE:
                    self.summary_list.append(tf.summary.histogram("z_mu", self.z_mu[0]))
                    self.summary_list.append(tf.summary.histogram("z_var", self.z_var[0]))
                    self.summary_list.append(tf.summary.histogram("z_logvar", self.z_logvar[0]))

        ## Decoder
        self.h_dims_decoder = h_dims_decoder
        self.decoder_variable_scope = tf.get_variable_scope().name
        prob_layer = self.build_decoder(h_temp = self.z_sample,
                                        h_dims = h_dims_decoder,
                                        variable_scope = "Decoder",
                                        kernel_initializer = self.kernel_initializer,
                                        l2_scale = self.l2_scale,
                                        l1_scale = self.l1_scale)

        if self.tensorboard:
            with tf.name_scope("X"):
                self.summary_list.append(tf.summary.histogram("X_mu", self.X_mu[0]))
                if self.isVAE and self.decoder_var != 'deterministic':
                    self.summary_list.append(tf.summary.histogram("X_var", self.X_var))
                    self.summary_list.append(tf.summary.histogram("X_logvar", self.X_logvar))
                    self.summary_list.append(tf.summary.histogram("X_eps", self.X_eps[0]))

        return self.X_sample


    def initialize_feed_dict(self, feed_dict = {}):
        """ Create a new feed_dict from input feed_dict and self.feed_dict """
        feed_dict_new = {}
        feed_dict_new.update(self.feed_dict)
        feed_dict_new.update(feed_dict)
        return feed_dict_new


    def train(self, sess, dataset = None, initialize = True, close_writer = True,
              train_writer = None, test_writer = None, initial_iter = 0):

        logging.info('Start training')

        start_time = time.time()

        self.sess = sess

        # Set random seed
        if initialize:
            with tf.name_scope('initialize'):
                tf.set_random_seed(self.random_seed)
                tf.logging.set_verbosity(tf.logging.FATAL)
                sess.run(tf.global_variables_initializer())

        if dataset is None:
            dataset = self.data_handler.dataset

        train_data_full, test_data, y_train_full, y_test = dataset

        # Split into validation
        if self.validation_split == 0:
            train_data = train_data_full
            y_train = y_train_full
            validation_data = test_data
            y_validation = y_test
        else:
            len_train = int(train_data.shape[0])
            len_validate = int(len_train * self.validation_split)
            validation_data = train_data_full[:len_validate]
            train_data      = train_data_full[len_validate:]
            y_validation    = y_train[:len_validate]
            y_train         = y_train[len_validate:]

        logging.info("======================== Data Split =============================")
        logging.info("validation_data: " + str(validation_data.shape))
        logging.info("y_validation: " + str(y_validation.shape))
        logging.info("train_data: " + str(train_data.shape))
        logging.info("y_train: " + str(y_train.shape))
        logging.info(np.min(np.std(y_train,0)))
        logging.info("test_data: " + str(test_data.shape))
        logging.info("y_test: " + str(y_test.shape))
        logging.info(np.min(np.std(y_test,0)))

        if self.tensorboard:
            "Starting Tensorboard "
            self.start_tensorboard(train_writer = train_writer, test_writer=  test_writer)

        feed_dict             = self.initialize_feed_dict()

        feed_dict_train       = self.initialize_feed_dict({self.X: train_data,
                                                           self.X_target: y_train})

        feed_dict_test        = self.initialize_feed_dict({self.X: test_data,
                                                           self.X_target: y_test})

        feed_dict_validation  = self.initialize_feed_dict({self.X: validation_data,
                                                           self.X_target: y_validation})

        self.feed_dict_test       = feed_dict_test
        self.feed_dict_train      = feed_dict_train
        self.feed_dict_validation = feed_dict_validation

        if self.batch_norm:
            feed_dict[self.is_train] = True

        # Set up for early stopping
        it_tb = -1 + initial_iter
        it_test = 0
        it_min = 0
        min_loss = float("inf")
        result = None

        iter = self.iter
        if self.keep_prob is not None:
            feed_dict[self.prob] = self.keep_prob

        ## Set up minibatch
        # Set minibatch size
        if self.mb_size <= 1:
            mb_size = int(int(train_data.shape[0]) * self.mb_size)
        else:
            mb_size = int(self.mb_size)

        train_data_size = int(train_data.shape[0])
        epoch = int((iter * mb_size) / train_data_size) + 1

        if self.dataAPI:
            logging.info('Using dataAPI')
            self.dataset = tf.data.Dataset.from_tensor_slices(tuple(self.dataAPI_placeholders)).shuffle(train_data_size).repeat(epoch).batch(mb_size)
            self.initializer = self.iterator.make_initializer(self.dataset)
            sess.run(self.initializer, feed_dict = {self.dataAPI_placeholders[0]: train_data,
                                                    self.dataAPI_placeholders[1]: y_train})

        elif self.data_handler.iterator is not None:
            logging.info('Using data_handler iterator')

        else:
            logging.info('Using batch')
            batch = batch_handler(train_data_size, mb_size)

        # Set up batch
        self.solvers = [self.train_op]

        for it in range(iter):
            it_tb += 1
            it_test += 1

            if self.dataAPI:
                pass
            elif self.data_handler.iterator is not None:
                pass
            else:
                idx_out = batch.next_batch()
                feed_dict[self.X] = train_data.take(idx_out, axis = -2)
                feed_dict[self.X_target] =  y_train.take(idx_out, axis = -2)

            ## Train a batch
            results = self.sess.run(self.solvers + self.losses, feed_dict = feed_dict)

            # Log train/test results
            if (it_tb % self.log_frequency == 0) or (it_tb == iter - 1):
                logging.info('')
                logging.info('Iter: {}, time: {:.4}s'.format(it, time.time()- start_time))

                ## Batch Loss
                zipped = zip(self.losses_name, results[-len(self.losses_name):])
                format_args = [item for pack in zipped for item in pack]
                str_report = "Batch: {}={:.4}" + ", {}={:.4}" * (len(self.losses_name)-1)
                logging.info(str_report.format(*format_args))

                ## Train Loss
                results = sess.run(self.losses, feed_dict = feed_dict_train)
                results_train = results[-len(self.losses_name):]
                # Write to Summary (writing train data)
                if self.tensorboard:
                    summary = results.pop(0)
                    self.train_writer.add_summary(summary, str(it_tb))

                zipped = zip(self.losses_name,results[-len(self.losses_name):])
                format_args = [item for pack in zipped for item in pack]
                str_report = "Train: {}={:.4}" + ", {}={:.4}" * (len(self.losses_name)-1)
                logging.info(str_report.format(*format_args))

                ## Train Loss
                results = sess.run(self.losses, feed_dict = feed_dict_test)
                results_test = results[-len(self.losses_name):]

                zipped = zip(self.losses_name,results[-len(self.losses_name):])
                format_args = [item for pack in zipped for item in pack]
                str_report = "Test: {}={:.4}" + ", {}={:.4}" * (len(self.losses_name)-1)
                logging.info(str_report.format(*format_args))

                ## Validation loss
                if self.tensorboard:
                    run_args = {}
                    if self.metadata:
                        run_args["options"] = tf.RunOptions(trace_level=tf.RunOptions.FULL_TRACE)
                        run_args['run_metadata'] = tf.RunMetadata()
                    results = self.sess.run(self.losses, feed_dict = feed_dict_validation, **run_args)
                    summary = results.pop(0)
                    if self.metadata:
                        self.test_writer.add_run_metadata(run_args["run_metadata"], str(it_tb))
                    self.test_writer.add_summary(summary, str(it_tb))

                    zipped = zip(self.losses_name,results[-len(self.losses_name):])
                    format_args = [item for pack in zipped for item in pack]
                    str_report = "Validation: {}={:.4}" + ", {}={:.4}" * (len(self.losses_name)-1)
                    logging.info(str_report.format(*format_args))

                else:
                    results = sess.run(self.losses, feed_dict = feed_dict_validation)
                    zipped = zip(self.losses_name,results[-len(self.losses_name):])
                    format_args = [item for pack in zipped for item in pack]
                    str_report = "Validation: {}={:.4}" + ", {}={:.4}" * (len(self.losses_name)-1)
                    logging.info(str_report.format(*format_args))

                if self.early_stopping > 0:
                    logging.info("new min loss: {} vs old {}".format(results[-3], min_loss))
                    delta = min_loss - results[-3]
                    if delta > self.tolerance:
                        it_min = it
                        it_test = 0
                        min_loss =  results[-3]

                    elif it_test > self.early_stopping and it_tb > self.iter_min_early_stopping:
                        break

        if result is None:

            result = {}

            logging.info('y_test shape: {}'.format(y_test.shape))
            logging.info('feed_dict_test: {}'.format(feed_dict_test))

        output = {'model_evaluation': result}
        output['losses'] = {'train': results_train,
                            'test' : results_test,
                            'name' : self.losses_name}

        if self.tensorboard:
            logging.info("Reults saved to tensorboard")

        #### Calculate layers for full data and save results
        output['split_index'] = self.data_handler.split_idx
        output['labels']      = np.array(self.data_handler.X.obs_names)

        full_data = self.data_handler.X.X
        feed_dict_full = self.initialize_feed_dict({self.X: full_data})
        y_full = self.data_handler.y.X

        if self.save_recon:

            logging.info("y_full: {}".format(y_full.shape))

            y_reconstruct = np.array(self.reconstruct_X(feed_dict_full))
            y_residual = y_full - y_reconstruct
            y_result = np.array([y_full, y_reconstruct, y_residual])

            logging.info("y_result.shape = {}".format(y_result.shape))

            output['reconstruction'] = y_result

        if self.save_LE:

            if np.all(y_test == y_train):
                y_full = y_train
                full_data = train_data
            else:
                y_full = np.concatenate([y_test, y_train], axis = -2)
                full_data = np.concatenate([test_data, train_data], axis = -2)

            z_mu = np.array(self.calculate_z_mu(feed_dict_full))
            logging.info("Saving bottleneck layer")
            logging.info("=======================")
            logging.info("bottleneck.shape: {}".format(z_mu.shape))
            output['z_mu'] = z_mu

            y = np.array(self.calculate_y_mu(feed_dict_full))
            logging.info("Saving y layer")
            logging.info("=======================")
            logging.info("y_mu.shape: {}".format(y.shape))
            output['y_mu'] = y

            decoder_layers = self.calculate_decoder_layers(feed_dict_full)
            logging.info("Saving decoder layers")
            output['decoder_layers'] = decoder_layers

            if self.isVAE and self.decoder_var != 'deterministic':
                z_var = np.array(self.calculate_latent_variance(feed_dict_full))
                logging.info("Saving z_var")
                logging.info("=======================")
                logging.info("self.z_logvar: {}".format(z_var.shape))
                output['z_var'] = z_var

                X_var = np.array(self.calculate_decoder_variance(feed_dict_full))
                logging.info("Saving X_var")
                logging.info("=======================")
                logging.info("self.X_var: {}".format(X_var.shape))
                output['X_var'] = X_var

        if self.save_W:
            W = self.get_W()
            output['W'] = W

        if self.tensorboard and close_writer:
            self.train_writer.close()
            self.test_writer.close()

        return output


    def start_tensorboard(self, train_writer = None, test_writer = None):
        """ Start tensorboard """

        if train_writer is not None and test_writer is not None:
            self.train_writer = train_writer
            self.test_writer  = test_writer
        else:
            logdir_train = os.path.join(self.logdir_tf,self.name,'train')
            logdir_test  = os.path.join(self.logdir_tf,self.name,'test')
            self.train_writer = tf.summary.FileWriter(logdir_train, self.sess.graph)
            self.test_writer = tf.summary.FileWriter(logdir_test, self.sess.graph)
            logging.info("Starting Tensorboard: {}".format(logdir_train))


    def calculate_X_reconstruct(self, feed_dict_test):
        """ Calcualte y from input data """

        y_reconstruct = np.array(self.sess.run([self.X_out], feed_dict = feed_dict_test))

        logging.info("y_reconstruct shape: {}".format(np.array(y_reconstruct).shape))

        return y_reconstruct


    def reconstruct_X(self, feed_dict_in):
        """
        Calculate the X_reconstruct from the mean of the latent variables rather than sampled
        """

        y_mu = self.calculate_y_mu(feed_dict_in)

        feed_dict_Wy = self.initialize_feed_dict({self.y_sample: y_mu})

        feed_dict_Wy = {self.y_sample: y_mu}

        X_reconstruct = np.array(self.sess.run(self.X_dist.mu, feed_dict = feed_dict_Wy))

        logging.info("X_reconstruct shape: {}".format(np.array(X_reconstruct).shape))

        return X_reconstruct


    def calculate_y_sample(self, feed_dict_test):
        """
        Calcualte y from input data and return matrix with dimensions [n_cell x n_gene]
        """

        y_reconstruct = np.array(self.sess.run(self.y_sample, feed_dict = feed_dict_test))

        logging.info("y_reconstruct shape: {}".format(np.array(y_reconstruct).shape))

        return y_reconstruct


    def calculate_z(self, feed_dict_input, sample = False):
        """ Calculate predicted mean of latent variable z from input data"""

        if sample:
            z = np.array(self.sess.run(self.z_sample, feed_dict = feed_dict_input))
        else:
            z = np.array(self.sess.run(self.z_mu, feed_dict = feed_dict_input))
        logging.info("z shape: {}".format(np.array(z).shape))

        return z


    def calculate_z_mu(self, feed_dict_input):
        """ Calculate predicted mean of latent variable z from input data"""

        if self.z_mu in feed_dict_input.keys():
            z_mu = feed_dict_input[self.z_mu]
        elif self.z_sample in feed_dict_input.keys():
            z_mu = feed_dict_input[self.z_sample]
        else:
            z_mu = np.array(self.sess.run(self.z_mu, feed_dict = feed_dict_input))

        logging.info("z_mu shape: {}".format(np.array(z_mu).shape))

        return z_mu


    def calculate_y_mu(self, feed_dict_input):
        """ Calculate predicted mean of latent variable y from input data without sampling z """

        z_mu = self.calculate_z_mu(feed_dict_input)

        if self.z_mu == self.y_mu:
            y_mu = z_mu
        else:
            y_mu = np.array(self.sess.run(self.y_mu, feed_dict = {self.z_sample: z_mu}))

        logging.info("y_mu: {}".format(np.array(y_mu).shape))

        return y_mu


    def calculate_decoder_layers(self, feed_dict_input, sample_z = False):
        """ Calculate predicted mean of latent variable y from input data without sampling z """

        z = self.calculate_z(feed_dict_input, sample = sample_z)

        hl_list = self.sess.run(self.decoder_layers, feed_dict = {self.z_sample: z})

        return hl_list


    def calculate_latent_variance(self, feed_dict_test):
        """ Calculate bottleneck layer from input data"""
        bottleneck = np.array(self.sess.run(self.z_var, feed_dict = feed_dict_test))

        logging.info("bottleneck shape: {}".format(np.array(bottleneck).shape))

        return bottleneck


    def calculate_decoder_variance(self, feed_dict_test):
        """ Calculate bottleneck layer from input data"""
        var = np.array(self.sess.run(self.X_var, feed_dict = feed_dict_test))

        logging.info("decoder variance shape: {}".format(np.array(var).shape))

        return var


    def get_W(self):
        """ Get W """
        W = np.array(self.sess.run(self.W))

        logging.info("W shape: {}".format(W.shape))

        return W


    def update_feed_dict(self, feed_dict):
        """ Update the basic feed_dict """
        self.feed_dict.update(feed_dict)
